
"""
======================================================================
 Project: Gedcom-X
 File:    mutations.py
 Author:  David J. Cartwright
 Purpose: Objects used to convert TAGs/Structues/Types from GEDCOM Versions
    when simple parsing will not work. (complex or ambiguous structures)

 Created: 2025-08-25
 Updated:
   - 2025-08-31: cleaned up imports and documentation
   - 2025-09-01: filename PEP8 standard, imports changed accordingly
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .gedcom5x import Gedcom5xRecord
from .fact import Fact, FactType
from .event import Event, EventType
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================

fact_event_table = {
    # Person Fact / Event Types
    "ADOP": {
        "Fact": FactType.AdoptiveParent,
        "Event": EventType.Adoption,
    },
    "CHR": {
        "Fact": FactType.AdultChristening,
        "Event": EventType.AdultChristening,
    },
    "EVEN": {
        "Fact": FactType.Amnesty,
        # no Event
    },
    "BAPM": {
        "Fact": FactType.Baptism,
        "Event": EventType.Baptism,
    },
    "BARM": {
        "Fact": FactType.BarMitzvah,
        "Event": EventType.BarMitzvah,
    },
    "BASM": {
        "Fact": FactType.BatMitzvah,
        "Event": EventType.BatMitzvah,
    },
    "BIRT": {
        "Fact": FactType.Birth,
        "Event": EventType.Birth,
    },
    "BIRT, CHR": {
        "Fact": FactType.Birth,
        "Event": EventType.Birth,
    },
    "BLES": {
        "Fact": FactType.Blessing,
        "Event": EventType.Blessing,
    },
    "BURI": {
        "Fact": FactType.Burial,
        "Event": EventType.Burial,
    },
    "CAST": {
        "Fact": FactType.Caste,
        # no Event
    },
    "CENS": {
        "Fact": FactType.Census,
        "Event": EventType.Census,
    },
    "CIRC": {
        "Fact": FactType.Circumcision,
        "Event": EventType.Circumcision,
    },
    "CONF": {
        "Fact": FactType.Confirmation,
        "Event": EventType.Confirmation,
    },
    "CREM": {
        "Fact": FactType.Cremation,
        "Event": EventType.Cremation,
    },
    "DEAT": {
        "Fact": FactType.Death,
        "Event": EventType.Death,
    },
    "EDUC": {
        "Fact": FactType.Education,
        "Event": EventType.Education,
    },
    "EMIG": {
        "Fact": FactType.Emigration,
        "Event": EventType.Emigration,
    },
    "FCOM": {
        "Fact": FactType.FirstCommunion,
        "Event": EventType.FirstCommunion,
    },
    "GRAD": {
        "Fact": FactType.Graduation,
        # no Event
    },
    "IMMI": {
        "Fact": FactType.Immigration,
        "Event": EventType.Immigration,
    },
    "MIL": {
        "Fact": FactType.MilitaryService,
        # no Event
    },
    "NATI": {
        "Fact": FactType.Nationality,
        # no Event
    },
    "NATU": {
        "Fact": FactType.Naturalization,
        "Event": EventType.Naturalization,
    },
    "OCCU": {
        "Fact": FactType.Occupation,
        # no Event
    },
    "ORDN": {
        "Fact": FactType.Ordination,
        "Event": EventType.Ordination,
    },
    "DSCR": {
        "Fact": FactType.PhysicalDescription,
        # no Event
    },
    "PROB": {
        "Fact": FactType.Probate,
        # no Event
    },
    "PROP": {
        "Fact": FactType.Property,
        # no Event
    },
    "RELI": {
        "Fact": FactType.Religion,
        # no Event
    },
    "RESI": {
        "Fact": FactType.Residence,
        # no Event
    },
    "WILL": {
        "Fact": FactType.Will,
        # no Event
    },

    # Couple Relationship Fact / Event Types
    "ANUL": {
        "Fact": FactType.Annulment,
        "Event": EventType.Annulment,
    },
    "DIV": {
        "Fact": FactType.Divorce,
        "Event": EventType.Divorce,
    },
    "DIVF": {
        "Fact": FactType.DivorceFiling,
        "Event": EventType.DivorceFiling,
    },
    "ENGA": {
        "Fact": FactType.Engagement,
        "Event": EventType.Engagement,
    },
    "MARR": {
        "Fact": FactType.Marriage,
        "Event": EventType.Marriage,
    },
    "MARB": {
        "Fact": FactType.MarriageBanns,
        # no Event
    },
    "MARC": {
        "Fact": FactType.MarriageContract,
        # no Event
    },
    "MARL": {
        "Fact": FactType.MarriageLicense,
        # no Event
    },
    "MARS":{
        "Fact":EventType.MarriageSettlment

    },
    "SEPA": {
        "Fact": FactType.Separation,
        # no Event
    },

}

class GedcomXObject:
    def __init__(self,record: Gedcom5xRecord) -> None:
        self.record = record
        self.created_with_tag: str | None = record.tag if record and isinstance(record, Gedcom5xRecord) else None    
        self.created_at_level: int | None = record.level if record and isinstance(record, Gedcom5xRecord) else None
        self.created_at_line_number: int | None = record.line if record and isinstance(record, Gedcom5xRecord) else None

class GedcomXSourceOrDocument(GedcomXObject):
    def __init__(self,record: Gedcom5xRecord) -> None:
        super().__init__(record)
        self.title: str | None = None
        self.citation: str | None = None
        self.page: str | None = None
        self.contributor: str | None = None
        self.publisher: str | None = None
        self.rights: str | None = None
        self.url: str | None = None
        self.medium: str | None = None
        self.type: str | None = None
        self.format: str | None = None
        self.created: str | None = None
        self.modified: str | None = None
        self.language: str | None = None
        self.relation: str | None = None
        self.identifier: str | None = None
        self.description: str | None = None

class GedcomXEventOrFact(GedcomXObject):
    def __new__(cls,record: Gedcom5xRecord, object_stack: dict | None = None) -> object:
        super().__init__(record)
        if record.tag in fact_event_table.keys():

            if 'Fact' in fact_event_table[record.tag].keys():
                obj = Fact(type=fact_event_table[record.tag]['Fact'])
                return obj
            elif 'Event' in fact_event_table[record.tag].keys():
                obj = Event(type=fact_event_table[record.tag]['Fact'])
            else:
                raise ValueError
        else:
            raise ValueError(f"{record.tag} not found in map")

class GedcomXRelationshipBuilder(GedcomXObject):
    def __init__(self, record: Gedcom5xRecord) -> None:
        super().__init__(record)

        




import warnings

from typing import Any, Optional, List

"""
======================================================================
 Project: Gedcom-X
 File:    conclusion.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-08-25
 Updated:
   - 2025-09-03: _from_json_ refactor 
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Type Imports
======================================================================
"""
from .attribution import Attribution
from .Extensions.rs10.rsLink import _rsLinks, rsLink
from .identifier import make_uid
from .note import Note
from .qualifier import Qualifier
from .resource import Resource, URI
from .source_reference import SourceReference
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================




class ConfidenceLevel(Qualifier):
    High = "http://gedcomx.org/High"
    Medium = "http://gedcomx.org/Medium"
    Low = "http://gedcomx.org/Low"

    _NAME_TO_URI = {
        "high": High,
        "medium": Medium,
        "low": Low,
    }

    @classmethod
    def _from_json_(cls, data):
        """
        Accepts:
          - "High" | "Medium" | "Low"
          - "http://gedcomx.org/High" | ".../Medium" | ".../Low"
          - {"type": "..."} or {"value": "..."} or {"confidence": "..."} or {"level": "..."} or {"uri": "..."}
          - existing ConfidenceLevel instance
        Returns:
          ConfidenceLevel instance with .value set to the canonical URI.
        """
        if data is None:
            return None

        if isinstance(data, cls):
            return data

        # Extract token from dicts or use the raw scalar
        if isinstance(data, dict):
            token = (
                data.get("confidence")
                or data.get("type")
                or data.get("value")
                or data.get("level")
                or data.get("uri")
            )
        else:
            token = data

        if token is None:
            return None

        token_str = str(token).strip()

        # Normalize to canonical URI
        if token_str.lower() in cls._NAME_TO_URI:
            uri = cls._NAME_TO_URI[token_str.lower()]
        elif token_str in (cls.High, cls.Medium, cls.Low):
            uri = token_str
        else:
            raise ValueError(f"Unknown ConfidenceLevel: {token!r}")

        # Create a ConfidenceLevel instance without invoking Qualifier.__init__
        obj = cls.__new__(cls)
        # store the canonical URI on the instance; used by description and (optionally) serialization
        obj.value = uri
        return obj

    @property
    def description(self):
        descriptions = {
            self.High: "The contributor has a high degree of confidence that the assertion is true.",
            self.Medium: "The contributor has a medium degree of confidence that the assertion is true.",
            self.Low: "The contributor has a low degree of confidence that the assertion is true."
        }
        # Works whether the instance holds .value or (edge-case) if `self` is compared directly
        key = getattr(self, "value", self)
        return descriptions.get(key, "No description available.")

    
class Conclusion():
    """
    Represents a conclusion in the GEDCOM X conceptual model. A conclusion is a 
    genealogical assertion about a person, relationship, or event, derived from 
    one or more sources, with optional supporting metadata such as confidence, 
    attribution, and notes.

    Args:
        id (str, optional): A unique identifier for the conclusion. If not provided, 
            a UUID-based identifier will be automatically generated.
        lang (str, optional): The language code of the conclusion. 
        sources (list[SourceReference], optional): A list of source references that 
            support the conclusion.
        analysis (Document | Resource, optional): A reference to an analysis document 
            or resource that supports the conclusion.
        notes (list[Note], optional): A list of notes providing additional context. 
            Defaults to an empty list.
        confidence (ConfidenceLevel, optional): The contributor's confidence in the 
            conclusion (High, Medium, or Low).
        attribution (Attribution, optional): Information about who contributed the 
            conclusion and when.
        uri (Resource, optional): A URI reference for the conclusion. Defaults to a 
            URI with the fragment set to the `id`.
        links (_LinkList, optional): A list of links associated with the conclusion. 
            Defaults to an empty `_LinkList`.  
    """
    identifier = 'http://gedcomx.org/v1/Conclusion'
    version = 'http://gedcomx.org/conceptual-model/v1'
   
    def __init__(self,
                 id: Optional[str] = None,
                 lang: Optional[str] = None,
                 sources: Optional[List[SourceReference]] = None,
                 analysis: Optional[object | Resource] = None,
                 notes: Optional[List[Note]] = None,
                 confidence: Optional[ConfidenceLevel] = None,
                 attribution: Optional[Attribution] = None,
                 uri: Optional[Resource] = None,
                 _max_note_count: int = 20,
                 links: Optional[_rsLinks] = None) -> None:
        
        
        self.id = id if id else make_uid()
        self.lang = lang
        self.sources = sources if sources else []
        self.analysis = analysis
        self.notes = notes if notes else []
        self.confidence = confidence
        self.attribution = attribution
        self.max_note_count = _max_note_count
        self.links = links if links else _rsLinks()    #NOTE This is not in specification, following FS format
    
    def add_note(self,note_to_add: Note):
        if self.notes and len(self.notes) >= self.max_note_count:
            warnings.warn(f"Max not count of {self.max_note_count} reached for id: {self.id}")
            return False
        if note_to_add and isinstance(note_to_add,Note):
            for existing in self.notes:
                if note_to_add == existing:
                    return False
            self.notes.append(note_to_add)

    def add_source_reference(self, source_to_add: SourceReference):
        if source_to_add and isinstance(source_to_add,SourceReference):
            for current_source in self.sources:
                if source_to_add == current_source:
                    return
            self.sources.append(source_to_add)
        else:
            raise ValueError()
        
    def add_link(self,link: rsLink) -> bool:
        """
        Adds a link to the Conclusion link list.

        Args:
            link (rsLink): The link to be added.

        Returns:
            bool: The return value. True for success, False otherwise.
        
        Note: Duplicate checking not impimented at this level
        """
        if link and isinstance(link,rsLink):
            self.links.add(link)
            return True
        return False
    
    @property
    def _as_dict_(self):
        with hub.use(serial_log):
            log.debug(f"Serializing 'Conclusion' with id: {self.id}")
            type_as_dict = {}

            if self.id:
                type_as_dict['id'] = self.id
            if self.lang:
                type_as_dict['lang'] = self.lang
            if self.sources and self.sources != []:
                type_as_dict['sources'] = [s._as_dict_ for s in self.sources if s]
            if self.analysis:
                type_as_dict['analysis'] = getattr(self.analysis, '_as_dict_', self.analysis)
            if self.notes and self.notes != []:
                type_as_dict['notes'] = [
                    (n._as_dict_ if hasattr(n, '_as_dict_') else n) for n in self.notes if n
                ]
            if self.confidence is not None:
                type_as_dict['confidence'] = self.confidence
            if self.attribution is not None:
                type_as_dict['attribution'] = self.attribution._as_dict_ 
            if self.links is not None:
                print(self.links)
                type_as_dict['links'] = self.links._as_dict_
            log.debug(f"'Conclusion' serialized with fields: {type_as_dict.keys()}") 
            if type_as_dict == {}: log.warning("serializing and empty 'Conclusion'")
      
        return type_as_dict if type_as_dict != {} else None
        
    
    

    @classmethod
    def _dict_from_json_(cls, data: dict, context=None) -> dict:
        conclusion = {}

        # Scalars
        if (id_ := data.get("id")) is not None:
            conclusion["id"] = id_

        if (lang := data.get("lang")) is not None:
            conclusion["lang"] = lang

        # Lists
        if (sources := data.get("sources")) is not None:
            conclusion["sources"] = [
                SourceReference._from_json_(x, context) for x in sources
            ]

        if (notes := data.get("notes")) is not None:
            conclusion["notes"] = [
                Note._from_json_(x, context) for x in notes
            ]

        # Objects
        if (analysis := data.get("analysis")) is not None:
            # depending on your model, analysis might be a Resource or Document
            conclusion["analysis"] = Resource._from_json_(analysis, context)

        if (confidence := data.get("confidence")) is not None:
            conclusion["confidence"] = ConfidenceLevel._from_json_(confidence, context)

        if (attribution := data.get("attribution")) is not None:
            conclusion["attribution"] = Attribution._from_json_(attribution, context)

        if (uri := data.get("uri")) is not None:
            conclusion["uri"] = URI(uri)

        if (links := data.get("links")) is not None:
            conclusion["links"] = _rsLinks._from_json_(links, context)
            if isinstance(conclusion["links"],dict): assert False

        # Constant / defaults
        #conclusion["_max_note_count"] = 20

        #return cls(**conclusion)
        return conclusion


    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        
        return (
            self.id == other.id and
            self.lang == other.lang and
            self.sources == other.sources and
            self.analysis == other.analysis and
            self.notes == other.notes and
            self.confidence == other.confidence and
            self.attribution == other.attribution
        )


from typing import Optional
"""
======================================================================
 Project: Gedcom-X
 File:    coverage.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-08-25
 Updated:
   - 2025-09-03: _from_json_ refactor 
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .date import Date
from .place_reference import PlaceReference
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================


class Coverage:
    identifier = 'http://gedcomx.org/v1/Coverage'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self,spatial: Optional[PlaceReference], temporal: Optional[Date]) -> None:
        self.spatial = spatial
        self.temporal = temporal    
    
    # ...existing code...

    @property
    def _as_dict_(self):
        from .serialization import Serialization
        type_as_dict = {}
        if self.spatial:
            type_as_dict['spatial'] = getattr(self.spatial, '_as_dict_', self.spatial)
        if self.temporal:  # (fixed: no space after the dot)
            type_as_dict['temporal'] = getattr(self.temporal, '_as_dict_', self.temporal)
        return Serialization.serialize_dict(type_as_dict) 

    @classmethod
    def _from_json_(cls, data: dict):
        """
        Create a Coverage instance from a JSON-dict (already parsed).
        """
        from .place_reference import PlaceReference
        from .date import Date

        spatial = PlaceReference._from_json_(data.get('spatial')) if data.get('spatial') else None
        temporal = Date._from_json_(data.get('temporal')) if data.get('temporal') else None
        return cls(spatial=spatial, temporal=temporal)


from enum import Enum
from typing import Any, Dict, List, Optional
"""
======================================================================
 Project: Gedcom-X
 File:    document.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-08-25
 Updated:
   - 2025-09-03: _from_json_ refactored 
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .attribution import Attribution
from .conclusion import Conclusion, ConfidenceLevel
from .note import Note
from .resource import Resource
from .source_reference import SourceReference
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================


class DocumentType(Enum):
    Abstract = "http://gedcomx.org/Abstract"
    Transcription = "http://gedcomx.org/Transcription"
    Translation = "http://gedcomx.org/Translation"
    Analysis = "http://gedcomx.org/Analysis"
    
    @property
    def description(self):
        descriptions = {
            DocumentType.Abstract: "The document is an abstract of a record or document.",
            DocumentType.Transcription: "The document is a transcription of a record or document.",
            DocumentType.Translation: "The document is a translation of a record or document.",
            DocumentType.Analysis: "The document is an analysis done by a researcher; a genealogical proof statement is an example of one kind of analysis document."
        }
        return descriptions.get(self, "No description available.")

class TextType(Enum):
    plain = 'plain'
    xhtml = 'xhtml'

class Document(Conclusion):
    identifier = 'http://gedcomx.org/v1/Document'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self, id: Optional[str] = None,
                 lang: Optional[str] = None,
                 sources: Optional[List[SourceReference]] = None,
                 analysis: Optional[Resource] = None,
                 notes: Optional[List[Note]] = None,
                 confidence: Optional[ConfidenceLevel] = None, # ConfidenceLevel
                 attribution: Optional[Attribution] = None,
                 type: Optional[DocumentType] = None,
                 extracted: Optional[bool] = None, # Default to False
                 textType: Optional[TextType] = None,
                 text: Optional[str] = None,
                 ) -> None:
        super().__init__(id, lang, sources, analysis, notes, confidence, attribution)
        self.type = type
        self.extracted = extracted
        self.textType = textType
        self.text = text
    
    @property
    def _as_dict(self):
        from .serialization import Serialization
        type_as_dict = super()._as_dict_
        if self.type:
            type_as_dict['type'] = self.type.value
        if self.extracted is not None:
            type_as_dict['extracted'] = self.extracted
        if self.textType:
            type_as_dict['textType'] = self.textType.value
        if self.text:
            type_as_dict['text'] = self.text
        return Serialization.serialize_dict(type_as_dict)
    
    @classmethod
    def _from_json_(cls, data: Any, context: Any = None) -> "Document":
        """
        Build a Document from JSON.
        Shorthand: a bare string becomes {'text': <string>}.
        """
        if not isinstance(data, dict):
            raise TypeError(f"{cls.__name__}._from_json_ expected dict or str, got {type(data)}")

        obj: Dict[str, Any] = Conclusion._dict_from_json_(data,context)

        # type (enum)
        if (typ := data.get("type")) is not None:
            obj["type"] = DocumentType(typ)
            

        # extracted (bool; accept common string forms)
        if (ex := data.get("extracted")) is not None:
            obj["extracted"] = bool(ex)

        # textType (enum)
        if (tt := data.get("textType")) is not None:
            obj["textType"] = TextType(tt) 

        # text (string)
        if (tx := data.get("text")) is not None:
            obj["text"] = str(tx)

        return cls(**obj)



from __future__ import annotations
from typing import Any, Dict, List, Optional, Union, Iterable
from collections import defaultdict


from .GedcomStructure import GedcomStructure
from . import specification as g7specs
from .logger import get_logger


from typing import Dict, List, Optional



class Gedcom7:
    def __init__(self, filepath: Optional[str] = None):
        self.persons: List[Any] = []
        self.families: List[Any] = []
        self.sources: List[Any] = []
        self.records: List['GedcomStructure'] = []
        self._tag_index: Dict[str, List[int]] = defaultdict(list)  # tag -> list of record indices

    # ---- indexing helpers -------------------------------------------------
    @staticmethod
    def _norm_tag(tag: str) -> str:
        return tag.upper()

    def _rebuild_index(self) -> None:
        self._tag_index.clear()
        for i, rec in enumerate(self.records):
            if getattr(rec, "tag", None):
                self._tag_index[self._norm_tag(rec.tag)].append(i)

    # Optional: keep index in sync if you append records elsewhere
    def _append_record(self, rec: 'GedcomStructure') -> None:
        self.records.append(rec)
        if getattr(rec, "tag", None):
            self._tag_index[self._norm_tag(rec.tag)].append(len(self.records) - 1)

    # ---- Python container protocol ----------------------------------------
    def __len__(self) -> int:
        return len(self.records)

    def __iter__(self) -> Iterable['GedcomStructure']:
        return iter(self.records)

    def __contains__(self, key: Union[str, 'GedcomStructure']) -> bool:
        if isinstance(key, str):
            return self._norm_tag(key) in self._tag_index
        return key in self.records

    def __getitem__(self, key: Union[int, slice, str, tuple]) -> Union['GedcomStructure', List['GedcomStructure']]:
        # by position
        if isinstance(key, (int, slice)):
            return self.records[key]

        # by tag
        if isinstance(key, str):
            idxs = self._tag_index.get(self._norm_tag(key), [])
            return [self.records[i] for i in idxs]

        # combo: ('INDI', 0) or ('INDI', 0:5)
        if isinstance(key, tuple) and len(key) == 2 and isinstance(key[0], str):
            tag, sub = key
            items = self[tag]  # list for that tag
            if isinstance(sub, int) or isinstance(sub, slice):
                return items[sub]
            raise TypeError(f"Unsupported sub-key type: {type(sub)!r}")

        raise TypeError(f"Unsupported key type: {type(key)!r}")

    # ---- your existing methods (trimmed) ----------------------------------
    @staticmethod   
    def parse_gedcom_line(line: str) -> Optional[Dict[str, Any]]:
                
        line = line.lstrip('\ufeff').rstrip('\r\n')
        if not line:
            return None

        parts = line.split(maxsplit=3)
        if len(parts) < 2:
            return None  # not even "0 HEAD"

        # 1) Level
        try:
            level = int(parts[0])
        except ValueError:
            return None

        # 2) Is parts[1] an XREF?
        xref = None
        if parts[1].startswith('@') and parts[1].endswith('@'):
            xref = parts[1]

        # 3) Where is the tag?
        if xref:
            # must have at least ["0", "@X@", "TAG"]
            if len(parts) < 3:
                return None
            tag = parts[2]
            # everything after index 2 is the value
            value_parts = parts[3:]  # could be empty or one-element
        else:
            tag = parts[1]
            # everything after index 1 is the value
            value_parts = parts[2:]  # could be empty, one- or two-element
            

        # 4) re-assemble the full value
        value = " ".join(value_parts)  # empty string if value_parts == []
        if value.startswith('@') and value.endswith('@'):
            xref = parts[1]

        if tag == 'TAG':
            xtag, uri = value.split()
            g7specs.g7_structure_specs[xtag] = uri
            g7specs.g7_structure_specs[uri] = {'label': 'Extension_' + xtag}
            
        return {
            "level": level,
            "xref": xref,
            "tag": tag,
            "value": value
        }


    def loadfile(self, filepath: str) -> None:
        log = get_logger('importlog')
        context: Dict[int, GedcomStructure] = {}
        records: List[GedcomStructure] = []

        with open(filepath, 'r', encoding='utf8') as file:
            for lineno, raw in enumerate(file, start=1):
                record = Gedcom7.parse_gedcom_line(raw)
                if record is None:
                    log.error(f'empty line at {lineno}: {raw}')
                    continue

                level = int(record["level"])
                if record["tag"] == g7specs.CONT:
                    context[level - 1].value += "\n" + record["value"]
                    continue

                structure = GedcomStructure(
                    level=level,
                    tag=record["tag"],
                    xref=record["xref"],
                    text=record["value"],
                    parent=context[level - 1] if level > 0 else None,
                    line_num=lineno
                )

                if level == 0:
                    records.append(structure)

                context[level] = structure

        self.records = records
        self._rebuild_index()  # <-- build fast tag index once



from typing import Any, Dict, Optional, List

"""
======================================================================
 Project: Gedcom-X
 File:    address.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-08-25
 Updated:
   - 2025-09-03: _from_json_ refactoring
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================

class Address:
    """A GedcomX Address Data Type
    A GedcomX Address Data Type.

    Represents a postal address according to the GedcomX conceptual model.

    Args:
        value (str, optional): A complete address as a single string.
        city (str, optional): Name of the city or town.
        country (str, optional): Name of the country.
        postalCode (str, optional): Postal or ZIP code.
        stateOrProvince (str, optional): Name of the state, province, or region.
        street (str, optional): First street address line.
        street2 (str, optional): Second street address line.
        street3 (str, optional): Third street address line.
        street4 (str, optional): Fourth street address line.
        street5 (str, optional): Fifth street address line.
        street6 (str, optional): Sixth street address line.
    """

    identifier = "http://gedcomx.org/v1/Address"
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self, value: Optional[str] = None,
                 city: Optional[str] = None,
                 country: Optional[str] = None,
                 postalCode: Optional[str] = None,
                 stateOrProvince: Optional[str] = None,
                 street: Optional[str] = None,
                 street2: Optional[str] = None,
                 street3: Optional[str] = None,
                 street4: Optional[str] = None,
                 street5: Optional[str] = None,
                 street6: Optional[str] = None):
        
        self._value = value #TODO impliment a parser for date strings.
        self.city = city
        self.country = country
        self.postalCode = postalCode
        self.stateOrProvince = stateOrProvince
        self.street = street
        self.street2 = street2
        self.street3 = street3
        self.street4 = street4
        self.street5 = street5
        self.street6 = street6

    @property
    def value(self) -> str: 
        return ', '.join(filter(None, [
            self.street, self.street2, self.street3,
            self.street4, self.street5, self.street6,
            self.city, self.stateOrProvince,
            self.postalCode, self.country
        ]))
    
    @value.setter
    def value(self,value: str):
        self._value = value
        return
        raise NotImplementedError("Parsing of a full address is not implimented.")
    
    def _append(self,value):
        if self._value:
            self._value = self._value + ' ' + value
        else:
            self._value = value
             
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        
        return (
            self.value == other.value and
            self.city == other.city and
            self.country == other.country and
            self.postalCode == other.postalCode and
            self.stateOrProvince == other.stateOrProvince and
            self.street == other.street and
            self.street2 == other.street2 and
            self.street3 == other.street3 and
            self.street4 == other.street4 and
            self.street5 == other.street5 and
            self.street6 == other.street6
        )
    
    def __str__(self) -> str:
        # Combine non-empty address components into a formatted string
        parts = [
            self._value,
            self.street,
            self.street2,
            self.street3,
            self.street4,
            self.street5,
            self.street6,
            self.city,
            self.stateOrProvince,
            self.postalCode,
            self.country
        ]

        # Filter out any parts that are None or empty strings
        filtered_parts = [str(part) for part in parts if part]

        # Join the remaining parts with a comma and space
        return ', '.join(filtered_parts)
    
    @property
    def _as_dict_(self):
        with hub.use(serial_log):
            log.debug(f"Serializing 'Address' with value: '{self.value}'")
            type_as_dict = {}
            if self.city: type_as_dict["city"] = self.city
            if self.country: type_as_dict["country"] = self.country
            if self.postalCode: type_as_dict["postalCode"] = self.postalCode
            if self.stateOrProvince: type_as_dict["stateOrProvince"] = self.stateOrProvince
            if self.street: type_as_dict["street"] = self.street
            if self.street2: type_as_dict["street2"] = self.street2
            if self.street3: type_as_dict["street3"] = self.street3
            if self.street4: type_as_dict["street4"] = self.street4
            if self.street5: type_as_dict["street5"] = self.street5
            if self.street6: type_as_dict["street6"] = self.street6
            log.debug(f"'Address' serialized with fields: '{type_as_dict.keys()}'") 
            if type_as_dict == {} or len(type_as_dict.keys()) == 0: log.warning("serializing and empty 'Address' Object")

        return type_as_dict if type_as_dict != {} else None
        
    
    @classmethod
    def _from_json_(cls, data: Any, context: Any = None) -> "Address":
        """
        Build an Address from JSON.
        Supports:
          - Shorthand string -> value
          - Aliases: postal_code/postal -> postalCode; state/province -> stateOrProvince
          - Line aliases: line1..line6 / address1..address6 / addr1..addr6 -> street..street6
          - 'lines': [..] list -> street..street6
        """
        if data is None: return None
        
        if not isinstance(data, dict):
            raise TypeError(f"{cls.__name__}._from_json_ expected dict or str, got {type(data)}")

        address_data: Dict[str, Any] = {}

        # Freeform value (accept a few aliases)
        if (v := data.get("value")) is None:
            address_data["value"] = str(v)

        # Simple scalars
        if (city := data.get("city")) is not None:
            address_data["city"] = city
        if (country := data.get("country")) is not None:
            address_data["country"] = country

        # Postal code (aliases)
        if (postal := data.get("postalCode")) is not None:
            address_data["postalCode"] = postal

        # State / Province (aliases)
        if (stateprov := data.get("stateOrProvince")) is not None:
            address_data["stateOrProvince"] = stateprov

        if data.get("street") is not None:
            address_data["street"] = data["street"]

        if data.get("street2") is not None:
            address_data["street2"] = data["street2"]

        if data.get("street3") is not None:
            address_data["street3"] = data["street3"]

        if data.get("street4") is not None:
            address_data["street4"] = data["street4"]

        if data.get("street5") is not None:
            address_data["street5"] = data["street5"]

        if data.get("street6") is not None:
            address_data["street6"] = data["street6"]


        return cls(**address_data)





from .. import *

g7toX = {
  "ABBR": "https://gedcom.io/terms/v7/ABBR",
  "ADDR": "https://gedcom.io/terms/v7/ADDR",
  "ADOP": "https://gedcom.io/terms/v7/ADOP",
  "ADOP-FAMC": "https://gedcom.io/terms/v7/ADOP-FAMC",
  "ADR1": "https://gedcom.io/terms/v7/ADR1",
  "ADR2": "https://gedcom.io/terms/v7/ADR2",
  "ADR3": "https://gedcom.io/terms/v7/ADR3",
  "AGE": "https://gedcom.io/terms/v7/AGE",
  "AGNC": "https://gedcom.io/terms/v7/AGNC",
  "ALIA": "https://gedcom.io/terms/v7/ALIA",
  "ANCI": "https://gedcom.io/terms/v7/ANCI",
  "ANUL": "https://gedcom.io/terms/v7/ANUL",
  "ASSO": "https://gedcom.io/terms/v7/ASSO",
  "AUTH": "https://gedcom.io/terms/v7/AUTH",
  "BAPL": "https://gedcom.io/terms/v7/BAPL",
  "BAPM": "https://gedcom.io/terms/v7/BAPM",
  "BARM": "https://gedcom.io/terms/v7/BARM",
  "BASM": "https://gedcom.io/terms/v7/BASM",
  "BIRT": "https://gedcom.io/terms/v7/BIRT",
  "BLES": "https://gedcom.io/terms/v7/BLES",
  "BURI": "https://gedcom.io/terms/v7/BURI",
  "CALN": "https://gedcom.io/terms/v7/CALN",
  "CAST": "https://gedcom.io/terms/v7/CAST",
  "CAUS": "https://gedcom.io/terms/v7/CAUS",
  "CENS": "https://gedcom.io/terms/v7/CENS",
  "CHAN": "https://gedcom.io/terms/v7/CHAN",
  "CHIL": "https://gedcom.io/terms/v7/CHIL",
  "CHR": "https://gedcom.io/terms/v7/CHR",
  "CHRA": "https://gedcom.io/terms/v7/CHRA",
  "CITY": "https://gedcom.io/terms/v7/CITY",
  "CONF": "https://gedcom.io/terms/v7/CONF",
  "CONL": "https://gedcom.io/terms/v7/CONL",
  "CONT": "https://gedcom.io/terms/v7/CONT",
  "COPR": "https://gedcom.io/terms/v7/COPR",
  "CORP": "https://gedcom.io/terms/v7/CORP",
  "CREA": "https://gedcom.io/terms/v7/CREA",
  "CREM": "https://gedcom.io/terms/v7/CREM",
  "CROP": "https://gedcom.io/terms/v7/CROP",
  "CTRY": "https://gedcom.io/terms/v7/CTRY",
  "DATA": "https://gedcom.io/terms/v7/DATA",
  "DATA-EVEN": "https://gedcom.io/terms/v7/DATA-EVEN",
  "DATA-EVEN-DATE": "https://gedcom.io/terms/v7/DATA-EVEN-DATE",
  "DATE": "https://gedcom.io/terms/v7/DATE",
  "DATE-exact": "https://gedcom.io/terms/v7/DATE-exact",
  "DEAT": "https://gedcom.io/terms/v7/DEAT",
  "DESI": "https://gedcom.io/terms/v7/DESI",
  "DEST": "https://gedcom.io/terms/v7/DEST",
  "DIV": "https://gedcom.io/terms/v7/DIV",
  "DIVF": "https://gedcom.io/terms/v7/DIVF",
  "DSCR": "https://gedcom.io/terms/v7/DSCR",
  "EDUC": "https://gedcom.io/terms/v7/EDUC",
  "EMAIL": "https://gedcom.io/terms/v7/EMAIL",
  "EMIG": "https://gedcom.io/terms/v7/EMIG",
  "ENDL": "https://gedcom.io/terms/v7/ENDL",
  "ENGA": "https://gedcom.io/terms/v7/ENGA",
  "EVEN": "https://gedcom.io/terms/v7/EVEN",
  "EXID": "https://gedcom.io/terms/v7/EXID",
  "EXID-TYPE": "https://gedcom.io/terms/v7/EXID-TYPE",
  "FACT": "https://gedcom.io/terms/v7/FACT",
  "FAM": "https://gedcom.io/terms/v7/FAM",
  "FAM-CENS": "https://gedcom.io/terms/v7/FAM-CENS",
  "FAM-EVEN": "https://gedcom.io/terms/v7/FAM-EVEN",
  "FAM-FACT": "https://gedcom.io/terms/v7/FAM-FACT",
  "FAM-HUSB": "https://gedcom.io/terms/v7/FAM-HUSB",
  "FAM-NCHI": "https://gedcom.io/terms/v7/FAM-NCHI",
  "FAM-RESI": "https://gedcom.io/terms/v7/FAM-RESI",
  "FAM-WIFE": "https://gedcom.io/terms/v7/FAM-WIFE",
  "FAMC": "https://gedcom.io/terms/v7/FAMC",
  "FAMC-ADOP": "https://gedcom.io/terms/v7/FAMC-ADOP",
  "FAMC-STAT": "https://gedcom.io/terms/v7/FAMC-STAT",
  "FAMS": "https://gedcom.io/terms/v7/FAMS",
  "FAX": "https://gedcom.io/terms/v7/FAX",
  "FCOM": "https://gedcom.io/terms/v7/FCOM",
  "FILE": "https://gedcom.io/terms/v7/FILE",
  "FILE-TRAN": "https://gedcom.io/terms/v7/FILE-TRAN",
  "FORM": "https://gedcom.io/terms/v7/FORM",
  "GEDC": "https://gedcom.io/terms/v7/GEDC",
  "GEDC-VERS": "https://gedcom.io/terms/v7/GEDC-VERS",
  "GIVN": "https://gedcom.io/terms/v7/GIVN",
  "GRAD": "https://gedcom.io/terms/v7/GRAD",
  "HEAD": "https://gedcom.io/terms/v7/HEAD",
  "HEAD-DATE": "https://gedcom.io/terms/v7/HEAD-DATE",
  "HEAD-LANG": "https://gedcom.io/terms/v7/HEAD-LANG",
  "HEAD-PLAC": "https://gedcom.io/terms/v7/HEAD-PLAC",
  "HEAD-PLAC-FORM": "https://gedcom.io/terms/v7/HEAD-PLAC-FORM",
  "HEAD-SOUR": "https://gedcom.io/terms/v7/HEAD-SOUR",
  "HEAD-SOUR-DATA": "https://gedcom.io/terms/v7/HEAD-SOUR-DATA",
  "HEIGHT": "https://gedcom.io/terms/v7/HEIGHT",
  "HUSB": "https://gedcom.io/terms/v7/HUSB",
  "IDNO": "https://gedcom.io/terms/v7/IDNO",
  "IMMI": "https://gedcom.io/terms/v7/IMMI",
  "INDI": "https://gedcom.io/terms/v7/INDI",
  "INDI-CENS": "https://gedcom.io/terms/v7/INDI-CENS",
  "INDI-EVEN": "https://gedcom.io/terms/v7/INDI-EVEN",
  "INDI-FACT": "https://gedcom.io/terms/v7/INDI-FACT",
  "INDI-FAMC": "https://gedcom.io/terms/v7/INDI-FAMC",
  "INDI-NAME": "https://gedcom.io/terms/v7/INDI-NAME",
  "INDI-NCHI": "https://gedcom.io/terms/v7/INDI-NCHI",
  "INDI-RELI": "https://gedcom.io/terms/v7/INDI-RELI",
  "INDI-RESI": "https://gedcom.io/terms/v7/INDI-RESI",
  "INDI-TITL": "https://gedcom.io/terms/v7/INDI-TITL",
  "INIL": "https://gedcom.io/terms/v7/INIL",
  "LANG": "https://gedcom.io/terms/v7/LANG",
  "LATI": "https://gedcom.io/terms/v7/LATI",
  "LEFT": "https://gedcom.io/terms/v7/LEFT",
  "LONG": "https://gedcom.io/terms/v7/LONG",
  "MAP": "https://gedcom.io/terms/v7/MAP",
  "MARB": "https://gedcom.io/terms/v7/MARB",
  "MARC": "https://gedcom.io/terms/v7/MARC",
  "MARL": "https://gedcom.io/terms/v7/MARL",
  "MARR": "https://gedcom.io/terms/v7/MARR",
  "MARS": "https://gedcom.io/terms/v7/MARS",
  "MEDI": "https://gedcom.io/terms/v7/MEDI",
  "MIME": "https://gedcom.io/terms/v7/MIME",
  "NAME": "https://gedcom.io/terms/v7/NAME",
  "NAME-TRAN": "https://gedcom.io/terms/v7/NAME-TRAN",
  "NAME-TYPE": "https://gedcom.io/terms/v7/NAME-TYPE",
  "NATI": "https://gedcom.io/terms/v7/NATI",
  "NATU": "https://gedcom.io/terms/v7/NATU",
  "NCHI": "https://gedcom.io/terms/v7/NCHI",
  "NICK": "https://gedcom.io/terms/v7/NICK",
  "NMR": "https://gedcom.io/terms/v7/NMR",
  "NO": "https://gedcom.io/terms/v7/NO",
  "NO-DATE": "https://gedcom.io/terms/v7/NO-DATE",
  "NOTE": "https://gedcom.io/terms/v7/NOTE",
  "NOTE-TRAN": "https://gedcom.io/terms/v7/NOTE-TRAN",
  "NPFX": "https://gedcom.io/terms/v7/NPFX",
  "NSFX": "https://gedcom.io/terms/v7/NSFX",
  "OBJE": "https://gedcom.io/terms/v7/OBJE",
  "OCCU": "https://gedcom.io/terms/v7/OCCU",
  "ORDN": "https://gedcom.io/terms/v7/ORDN",
  "PAGE": "https://gedcom.io/terms/v7/PAGE",
  "PEDI": "https://gedcom.io/terms/v7/PEDI",
  "PHON": "https://gedcom.io/terms/v7/PHON",
  "PHRASE": "https://gedcom.io/terms/v7/PHRASE",
  "PLAC": "https://gedcom.io/terms/v7/PLAC",
  "PLAC-FORM": "https://gedcom.io/terms/v7/PLAC-FORM",
  "PLAC-TRAN": "https://gedcom.io/terms/v7/PLAC-TRAN",
  "POST": "https://gedcom.io/terms/v7/POST",
  "PROB": "https://gedcom.io/terms/v7/PROB",
  "PROP": "https://gedcom.io/terms/v7/PROP",
  "PUBL": "https://gedcom.io/terms/v7/PUBL",
  "QUAY": "https://gedcom.io/terms/v7/QUAY",
  "REFN": "https://gedcom.io/terms/v7/REFN",
  "RELI": "https://gedcom.io/terms/v7/RELI",
  "REPO": "https://gedcom.io/terms/v7/REPO",
  "RESI": "https://gedcom.io/terms/v7/RESI",
  "RESN": "https://gedcom.io/terms/v7/RESN",
  "RETI": "https://gedcom.io/terms/v7/RETI",
  "ROLE": "https://gedcom.io/terms/v7/ROLE",
  "SCHMA": "https://gedcom.io/terms/v7/SCHMA",
  "SDATE": "https://gedcom.io/terms/v7/SDATE",
  "SEX": "https://gedcom.io/terms/v7/SEX",
  "SLGC": "https://gedcom.io/terms/v7/SLGC",
  "SLGS": "https://gedcom.io/terms/v7/SLGS",
  "SNOTE": "https://gedcom.io/terms/v7/SNOTE",
  "SOUR": "https://gedcom.io/terms/v7/SOUR",
  "SOUR-DATA": "https://gedcom.io/terms/v7/SOUR-DATA",
  "SOUR-EVEN": "https://gedcom.io/terms/v7/SOUR-EVEN",
  "SPFX": "https://gedcom.io/terms/v7/SPFX",
  "SSN": "https://gedcom.io/terms/v7/SSN",
  "STAE": "https://gedcom.io/terms/v7/STAE",
  "STAT": "https://gedcom.io/terms/v7/STAT",
  "SUBM": "https://gedcom.io/terms/v7/SUBM",
  "SUBM-LANG": "https://gedcom.io/terms/v7/SUBM-LANG",
  "SURN": "https://gedcom.io/terms/v7/SURN",
  "TAG": "https://gedcom.io/terms/v7/TAG",
  "TEMP": "https://gedcom.io/terms/v7/TEMP",
  "TEXT": "https://gedcom.io/terms/v7/TEXT",
  "TIME": "https://gedcom.io/terms/v7/TIME",
  "TITL": "https://gedcom.io/terms/v7/TITL",
  "TOP": "https://gedcom.io/terms/v7/TOP",
  "TRAN": "https://gedcom.io/terms/v7/TRAN",
  "TRLR": "https://gedcom.io/terms/v7/TRLR",
  "TYPE": "https://gedcom.io/terms/v7/TYPE",
  "UID": "https://gedcom.io/terms/v7/UID",
  "VERS": "https://gedcom.io/terms/v7/VERS",
  "WIDTH": "https://gedcom.io/terms/v7/WIDTH",
  "WIFE": "https://gedcom.io/terms/v7/WIFE",
  "WILL": "https://gedcom.io/terms/v7/WILL",
  "WWW": "https://gedcom.io/terms/v7/WWW",
  "enumset-ADOP": "https://gedcom.io/terms/v7/enumset-ADOP",
  "enumset-EVEN": "https://gedcom.io/terms/v7/enumset-EVEN",
  "enumset-EVENATTR": "https://gedcom.io/terms/v7/enumset-EVENATTR",
  "enumset-FAMC-STAT": "https://gedcom.io/terms/v7/enumset-FAMC-STAT",
  "enumset-MEDI": "https://gedcom.io/terms/v7/enumset-MEDI",
  "enumset-NAME-TYPE": "https://gedcom.io/terms/v7/enumset-NAME-TYPE",
  "enumset-PEDI": "https://gedcom.io/terms/v7/enumset-PEDI",
  "enumset-QUAY": "https://gedcom.io/terms/v7/enumset-QUAY",
  "enumset-RESN": "https://gedcom.io/terms/v7/enumset-RESN",
  "enumset-ROLE": "https://gedcom.io/terms/v7/enumset-ROLE",
  "enumset-SEX": "https://gedcom.io/terms/v7/enumset-SEX",
  "ord-STAT": "https://gedcom.io/terms/v7/ord-STAT",
  "record-FAM": "https://gedcom.io/terms/v7/record-FAM",
  "record-INDI": "https://gedcom.io/terms/v7/record-INDI",
  "record-OBJE": "https://gedcom.io/terms/v7/record-OBJE",
  "record-REPO": "https://gedcom.io/terms/v7/record-REPO",
  "record-SNOTE": "https://gedcom.io/terms/v7/record-SNOTE",
  "record-SOUR": "https://gedcom.io/terms/v7/record-SOUR",
  "record-SUBM": "https://gedcom.io/terms/v7/record-SUBM"
}




import difflib
import re

from enum import Enum
from typing import List, Optional, Dict, Any
"""
======================================================================
 Project: Gedcom-X
 File:    fact.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-08-25
 Updated:
   - 2025-09-03: _from_json_ refactor 
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .attribution import Attribution
from .conclusion import Conclusion, ConfidenceLevel
from .date import Date
from .document import Document
from .Extensions.rs10.rsLink import _rsLinks
from .note import Note
from .place_reference import PlaceReference
from .qualifier import Qualifier
from .resource import Resource
from .source_reference import SourceReference
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================


class FactType(Enum):
    # Person Fact Types
    Adoption = "http://gedcomx.org/Adoption"
    AdultChristening = "http://gedcomx.org/AdultChristening"
    Amnesty = "http://gedcomx.org/Amnesty"
    AncestralHall = "http://gedcomx.org/AncestralHall"
    AncestralPoem = "http://gedcomx.org/AncestralPoem"
    Apprenticeship = "http://gedcomx.org/Apprenticeship"
    Arrest = "http://gedcomx.org/Arrest"
    Award = "http://gedcomx.org/Award"
    Baptism = "http://gedcomx.org/Baptism"
    BarMitzvah = "http://gedcomx.org/BarMitzvah"
    BatMitzvah = "http://gedcomx.org/BatMitzvah"
    Birth = "http://gedcomx.org/Birth"
    BirthNotice = "http://gedcomx.org/BirthNotice"
    Blessing = "http://gedcomx.org/Blessing"
    Branch = "http://gedcomx.org/Branch"
    Burial = "http://gedcomx.org/Burial"
    Caste = "http://gedcomx.org/Caste"
    Census = "http://gedcomx.org/Census"
    Christening = "http://gedcomx.org/Christening"
    Circumcision = "http://gedcomx.org/Circumcision"
    Clan = "http://gedcomx.org/Clan"
    Confirmation = "http://gedcomx.org/Confirmation"
    Court = "http://gedcomx.org/Court"
    Cremation = "http://gedcomx.org/Cremation"
    Death = "http://gedcomx.org/Death"
    Education = "http://gedcomx.org/Education"
    EducationEnrollment = "http://gedcomx.org/EducationEnrollment"
    Emigration = "http://gedcomx.org/Emigration"
    Enslavement = "http://gedcomx.org/Enslavement"
    Ethnicity = "http://gedcomx.org/Ethnicity"
    Excommunication = "http://gedcomx.org/Excommunication"
    FirstCommunion = "http://gedcomx.org/FirstCommunion"
    Funeral = "http://gedcomx.org/Funeral"
    GenderChange = "http://gedcomx.org/GenderChange"
    GenerationNumber = "http://gedcomx.org/GenerationNumber"
    Graduation = "http://gedcomx.org/Graduation"
    Heimat = "http://gedcomx.org/Heimat"
    Immigration = "http://gedcomx.org/Immigration"
    Imprisonment = "http://gedcomx.org/Imprisonment"
    Inquest = "http://gedcomx.org/Inquest"
    LandTransaction = "http://gedcomx.org/LandTransaction"
    Language = "http://gedcomx.org/Language"
    Living = "http://gedcomx.org/Living"
    MaritalStatus = "http://gedcomx.org/MaritalStatus"
    Medical = "http://gedcomx.org/Medical"
    MilitaryAward = "http://gedcomx.org/MilitaryAward"
    MilitaryDischarge = "http://gedcomx.org/MilitaryDischarge"
    MilitaryDraftRegistration = "http://gedcomx.org/MilitaryDraftRegistration"
    MilitaryInduction = "http://gedcomx.org/MilitaryInduction"
    MilitaryService = "http://gedcomx.org/MilitaryService"
    Mission = "http://gedcomx.org/Mission"
    MoveFrom = "http://gedcomx.org/MoveFrom"
    MoveTo = "http://gedcomx.org/MoveTo"
    MultipleBirth = "http://gedcomx.org/MultipleBirth"
    NationalId = "http://gedcomx.org/NationalId"
    Nationality = "http://gedcomx.org/Nationality"
    Naturalization = "http://gedcomx.org/Naturalization"
    NumberOfChildren = "http://gedcomx.org/NumberOfChildren"
    NumberOfMarriages = "http://gedcomx.org/NumberOfMarriages"
    Obituary = "http://gedcomx.org/Obituary"
    OfficialPosition = "http://gedcomx.org/OfficialPosition"
    Occupation = "http://gedcomx.org/Occupation"
    Ordination = "http://gedcomx.org/Ordination"
    Pardon = "http://gedcomx.org/Pardon"
    PhysicalDescription = "http://gedcomx.org/PhysicalDescription"
    Probate = "http://gedcomx.org/Probate"
    Property = "http://gedcomx.org/Property"
    Race = "http://gedcomx.org/Race"
    Religion = "http://gedcomx.org/Religion"
    Residence = "http://gedcomx.org/Residence"
    Retirement = "http://gedcomx.org/Retirement"
    Stillbirth = "http://gedcomx.org/Stillbirth"
    TaxAssessment = "http://gedcomx.org/TaxAssessment"
    Tribe = "http://gedcomx.org/Tribe"
    Will = "http://gedcomx.org/Will"
    Visit = "http://gedcomx.org/Visit"
    Yahrzeit = "http://gedcomx.org/Yahrzeit"

    # Couple Relationship Fact Types
    Annulment = "http://gedcomx.org/Annulment"
    CommonLawMarriage = "http://gedcomx.org/CommonLawMarriage"
    CivilUnion = "http://gedcomx.org/CivilUnion"
    Divorce = "http://gedcomx.org/Divorce"
    DivorceFiling = "http://gedcomx.org/DivorceFiling"
    DomesticPartnership = "http://gedcomx.org/DomesticPartnership"
    Engagement = "http://gedcomx.org/Engagement"
    Marriage = "http://gedcomx.org/Marriage"
    MarriageBanns = "http://gedcomx.org/MarriageBanns"
    MarriageContract = "http://gedcomx.org/MarriageContract"
    MarriageLicense = "http://gedcomx.org/MarriageLicense"
    MarriageNotice = "http://gedcomx.org/MarriageNotice"
    Separation = "http://gedcomx.org/Separation"

    # Parent-Child Relationship Fact Types
    AdoptiveParent = "http://gedcomx.org/AdoptiveParent"
    BiologicalParent = "http://gedcomx.org/BiologicalParent"
    ChildOrder = "http://gedcomx.org/ChildOrder"
    EnteringHeir = "http://gedcomx.org/EnteringHeir"
    ExitingHeir = "http://gedcomx.org/ExitingHeir"
    FosterParent = "http://gedcomx.org/FosterParent"
    GuardianParent = "http://gedcomx.org/GuardianParent"
    StepParent = "http://gedcomx.org/StepParent"
    SociologicalParent = "http://gedcomx.org/SociologicalParent"
    SurrogateParent = "http://gedcomx.org/SurrogateParent"
    Unknown = 'null'

    @classmethod
    def from_value(cls, value: str):
        for member in cls:
            if member.value == value:
                return member
        return FactType.Unknown
    
    @property
    def description(self):
        descriptions = {
            FactType.Adoption: "A fact of a person's adoption.",
            FactType.AdultChristening: "A fact of a person's christening or baptism as an adult.",
            FactType.Amnesty: "A fact of a person's amnesty.",
            FactType.AncestralHall: "A fact of a person's ancestral hall.",
            FactType.AncestralPoem: "A fact of a person's ancestral poem.",
            FactType.Apprenticeship: "A fact of a person's apprenticeship.",
            FactType.Arrest: "A fact of a person's arrest.",
            FactType.Award: "A fact of a person's award (medal, honor).",
            FactType.Baptism: "A fact of a person's baptism.",
            FactType.BarMitzvah: "A fact of a person's bar mitzvah.",
            FactType.BatMitzvah: "A fact of a person's bat mitzvah.",
            FactType.Birth: "A fact of a person's birth.",
            FactType.BirthNotice: "A fact of a person's birth notice.",
            FactType.Blessing: "A fact of an official blessing received by a person.",
            FactType.Branch: "A fact of a person's branch within an extended clan.",
            FactType.Burial: "A fact of the burial of a person's body after death.",
            FactType.Caste: "A fact of a person's caste.",
            FactType.Census: "A fact of a person's participation in a census.",
            FactType.Christening: "A fact of a person's christening at birth.",
            FactType.Circumcision: "A fact of a person's circumcision.",
            FactType.Clan: "A fact of a person's clan.",
            FactType.Confirmation: "A fact of a person's confirmation.",
            FactType.Court: "A fact of the appearance of a person in a court proceeding.",
            FactType.Cremation: "A fact of the cremation of a person's body.",
            FactType.Death: "A fact of the death of a person.",
            FactType.Education: "A fact of a person's education or educational achievement.",
            FactType.EducationEnrollment: "A fact of a person's enrollment in an educational program.",
            FactType.Emigration: "A fact of the emigration of a person.",
            FactType.Enslavement: "A fact of the enslavement of a person.",
            FactType.Ethnicity: "A fact of a person's ethnicity.",
            FactType.Excommunication: "A fact of a person's excommunication.",
            FactType.FirstCommunion: "A fact of a person's first communion.",
            FactType.Funeral: "A fact of a person's funeral.",
            FactType.GenderChange: "A fact of a person's gender change.",
            FactType.GenerationNumber: "A fact of a person's generation number.",
            FactType.Graduation: "A fact of a person's graduation.",
            FactType.Heimat: "A fact of a person's heimat.",
            FactType.Immigration: "A fact of a person's immigration.",
            FactType.Imprisonment: "A fact of a person's imprisonment.",
            FactType.Inquest: "A legal inquest, often after a suspicious death.",
            FactType.LandTransaction: "A fact of a land transaction by a person.",
            FactType.Language: "A fact of a language spoken by a person.",
            FactType.Living: "A fact of a record of a person's living for a period.",
            FactType.MaritalStatus: "A fact of a person's marital status.",
            FactType.Medical: "A fact of a person's medical record.",
            FactType.MilitaryAward: "A fact of a person's military award.",
            FactType.MilitaryDischarge: "A fact of a person's military discharge.",
            FactType.MilitaryDraftRegistration: "A fact of a person's draft registration.",
            FactType.MilitaryInduction: "A fact of a person's military induction.",
            FactType.MilitaryService: "A fact of a person's military service.",
            FactType.Mission: "A fact of a person's church mission.",
            FactType.MoveFrom: "A fact of a person's move from a location.",
            FactType.MoveTo: "A fact of a person's move to a new location.",
            FactType.MultipleBirth: "A fact of a person's birth as part of a multiple birth.",
            FactType.NationalId: "A fact of a person's national ID.",
            FactType.Nationality: "A fact of a person's nationality.",
            FactType.Naturalization: "A fact of a person's naturalization.",
            FactType.NumberOfChildren: "A fact of the number of children.",
            FactType.NumberOfMarriages: "A fact of a person's number of marriages.",
            FactType.Obituary: "A fact of a person's obituary.",
            FactType.OfficialPosition: "A fact of a person's official government position.",
            FactType.Occupation: "A fact of a person's occupation.",
            FactType.Ordination: "A fact of a person's ordination.",
            FactType.Pardon: "A fact of a person's legal pardon.",
            FactType.PhysicalDescription: "A fact of a person's physical description.",
            FactType.Probate: "A fact of a person's probate receipt.",
            FactType.Property: "A fact of a person's property.",
            FactType.Race: "A fact of a person's race.",
            FactType.Religion: "A fact of a person's religion.",
            FactType.Residence: "A fact of a person's residence.",
            FactType.Retirement: "A fact of a person's retirement.",
            FactType.Stillbirth: "A fact of a person's stillbirth.",
            FactType.TaxAssessment: "A fact of a person's tax assessment.",
            FactType.Tribe: "A fact of a person's tribe.",
            FactType.Will: "A fact of a person's will.",
            FactType.Visit: "A fact of a person's visit to a place.",
            FactType.Yahrzeit: "A fact of a person's yahrzeit.",
            FactType.AdoptiveParent: "A fact about an adoptive relationship.",
            FactType.BiologicalParent: "A fact about the biological relationship.",
            FactType.ChildOrder: "A fact about the child order.",
            FactType.EnteringHeir: "A fact about an entering heir relationship.",
            FactType.ExitingHeir: "A fact about an exiting heir relationship.",
            FactType.FosterParent: "A fact about a foster relationship.",
            FactType.GuardianParent: "A fact about legal guardianship.",
            FactType.StepParent: "A fact about the step relationship.",
            FactType.SociologicalParent: "A fact about a sociological relationship.",
            FactType.SurrogateParent: "A fact about a surrogate relationship.",
            FactType.Annulment: "A fact of an annulment of marriage.",
            FactType.CommonLawMarriage: "A fact of marriage by common law.",
            FactType.CivilUnion: "A fact of a civil union.",
            FactType.DomesticPartnership: "A fact of a domestic partnership.",
            FactType.Engagement: "A fact of an engagement to marry.",
            FactType.MarriageBanns: "A fact of marriage banns.",
            FactType.MarriageContract: "A fact of a marriage contract.",
            FactType.MarriageLicense: "A fact of a marriage license.",
            FactType.MarriageNotice: "A fact of a marriage notice.",
            FactType.Separation: "A fact of a couple's separation."
        }
        return descriptions.get(self, "No description available.")
    
    @staticmethod
    def guess(description):
        keywords_to_fact_type = {
            # Person Fact Types
            "adoption": FactType.Adoption,
            "adult christening": FactType.AdultChristening,
            "amnesty": FactType.Amnesty,
            "ancestral hall": FactType.AncestralHall,
            "ancestral poem": FactType.AncestralPoem,
            "apprenticeship": FactType.Apprenticeship,
            "arrest": FactType.Arrest,
            "award": FactType.Award,
            "baptism": FactType.Baptism,
            "bar mitzvah": FactType.BarMitzvah,
            "bat mitzvah": FactType.BatMitzvah,
            "birth": FactType.Birth,
            "birth notice": FactType.BirthNotice,
            "blessing": FactType.Blessing,
            "branch": FactType.Branch,
            "burial": FactType.Burial,
            "caste": FactType.Caste,
            "census": FactType.Census,
            "christening": FactType.Christening,
            "circumcision": FactType.Circumcision,
            "clan": FactType.Clan,
            "confirmation": FactType.Confirmation,
            "court": FactType.Court,
            "cremation": FactType.Cremation,
            "death": FactType.Death,
            "education": FactType.Education,
            "education enrollment": FactType.EducationEnrollment,
            "emigration": FactType.Emigration,
            "enslavement": FactType.Enslavement,
            "ethnicity": FactType.Ethnicity,
            "excommunication": FactType.Excommunication,
            "first communion": FactType.FirstCommunion,
            "funeral": FactType.Funeral,
            "gender change": FactType.GenderChange,
            "generation number": FactType.GenerationNumber,
            "graduation": FactType.Graduation,
            "heimat": FactType.Heimat,
            "immigration": FactType.Immigration,
            "imprisonment": FactType.Imprisonment,
            "inquest": FactType.Inquest,
            "land transaction": FactType.LandTransaction,
            "language": FactType.Language,
            "living": FactType.Living,
            "marital status": FactType.MaritalStatus,
            "medical": FactType.Medical,
            "military award": FactType.MilitaryAward,
            "military discharge": FactType.MilitaryDischarge,
            "military draft registration": FactType.MilitaryDraftRegistration,
            "military induction": FactType.MilitaryInduction,
            "military service": FactType.MilitaryService,
            "mission": FactType.Mission,
            "move from": FactType.MoveFrom,
            "move to": FactType.MoveTo,
            "multiple birth": FactType.MultipleBirth,
            "national id": FactType.NationalId,
            "nationality": FactType.Nationality,
            "naturalization": FactType.Naturalization,
            "number of children": FactType.NumberOfChildren,
            "number of marriages": FactType.NumberOfMarriages,
            "obituary": FactType.Obituary,
            "official position": FactType.OfficialPosition,
            "occupation": FactType.Occupation,
            "ordination": FactType.Ordination,
            "pardon": FactType.Pardon,
            "physical description": FactType.PhysicalDescription,
            "probate": FactType.Probate,
            "property": FactType.Property,
            "race": FactType.Race,
            "religion": FactType.Religion,
            "residence": FactType.Residence,
            "retirement": FactType.Retirement,
            "stillbirth": FactType.Stillbirth,
            "tax assessment": FactType.TaxAssessment,
            "tribe": FactType.Tribe,
            "will": FactType.Will,
            "visit": FactType.Visit,
            "yahrzeit": FactType.Yahrzeit,

            # Couple Relationship Fact Types
            "annulment": FactType.Annulment,
            "common law marriage": FactType.CommonLawMarriage,
            "civil union": FactType.CivilUnion,
            "divorce": FactType.Divorce,
            "divorce filing": FactType.DivorceFiling,
            "domestic partnership": FactType.DomesticPartnership,
            "engagement": FactType.Engagement,
            "marriage": FactType.Marriage,
            "marriage banns": FactType.MarriageBanns,
            "marriage contract": FactType.MarriageContract,
            "marriage license": FactType.MarriageLicense,
            "marriage notice": FactType.MarriageNotice,
            "couple number of children": FactType.NumberOfChildren,
            "separation": FactType.Separation,

            # Parent-Child Relationship Fact Types
            "adoptive parent": FactType.AdoptiveParent,
            "biological parent": FactType.BiologicalParent,
            "child order": FactType.ChildOrder,
            "entering heir": FactType.EnteringHeir,
            "exiting heir": FactType.ExitingHeir,
            "foster parent": FactType.FosterParent,
            "guardian parent": FactType.GuardianParent,
            "step parent": FactType.StepParent,
            "sociological parent": FactType.SociologicalParent,
            "surrogate parent": FactType.SurrogateParent
        }


        description_lower = description.lower()
    
        # Replace any non-alphanumeric characters with a space
        description_clean = re.sub(r'[^a-z0-9\s]', ' ', description_lower)
        
        # Get a list of words in the cleaned description
        words = description_clean.split()

        # Check for the best matching keyword in the description
        for word in words:
            matches = difflib.get_close_matches(word, keywords_to_fact_type.keys(), n=1, cutoff=0.8)
            if matches:
                return keywords_to_fact_type[matches[0]]
        return None

class FactQualifier(Enum):
    Age = "http://gedcomx.org/Age"
    Cause = "http://gedcomx.org/Cause"
    Religion = "http://gedcomx.org/Religion"
    Transport = "http://gedcomx.org/Transport"
    NonConsensual = "http://gedcomx.org/NonConsensual"
    
    @property
    def description(self):
        descriptions = {
            FactQualifier.Age: "The age of a person at the event described by the fact.",
            FactQualifier.Cause: "The cause of the fact, such as the cause of death.",
            FactQualifier.Religion: "The religion associated with a religious event such as a baptism or excommunication.",
            FactQualifier.Transport: "The name of the transport associated with an event that indicates a move.",
            FactQualifier.NonConsensual: "An indicator that the event occurred non-consensually, e.g., under enslavement."
        }
        return descriptions.get(self, "No description available.")

class Fact(Conclusion):
    identifier = 'http://gedcomx.org/v1/Fact'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self,
                 id: Optional[str] = None,
                 lang: Optional[str] = None,
                 sources: Optional[List[SourceReference]] = None,
                 analysis: Optional[Resource | Document] = None,
                 notes: Optional[List[Note]] = None,
                 confidence: Optional[ConfidenceLevel] = None,
                 attribution: Optional[Attribution] = None,
                 type: Optional[FactType] = None,
                 date: Optional[Date] = None,
                 place: Optional[PlaceReference] = None,
                 value: Optional[str] = None,
                 qualifiers: Optional[List[FactQualifier]] = None,
                 links: Optional[_rsLinks] = None):
        super().__init__(id, lang, sources, analysis, notes, confidence, attribution, links=links)
        self.type = type
        self.date = date
        self.place = place
        self.value = value
        self._qualifiers = qualifiers if qualifiers else []
    

    @property
    def qualifiers(self) -> List[FactQualifier]:
        return self._qualifiers # type: ignore

    @qualifiers.setter
    def qualifiers(self, value: List[FactQualifier]):
        if (not isinstance(value, list)) or (not all(isinstance(item, FactQualifier) for item in value)):
            raise ValueError("sources must be a list of GedcomRecord objects.")
        self._qualifiers.extend(value)

    @property
    def _as_dict_(self):
        '''
        Standard GedcomX Type JSON Serialization
        Returns: dict that contains only field for which the object has data in
        '''
        with hub.use(serial_log):
            log.debug(f"Serializing 'Fact' with id: {self.id}")
            type_as_dict = super()._as_dict_ 
            if type_as_dict is None:
                log.debug(f"Subject had no fields, creating new dict") 
                type_as_dict ={}
            # Only add Relationship-specific fields
            if self.type:
                type_as_dict['type'] = getattr(self.type, 'value', self.type)
            if self.date:
                type_as_dict['date'] = self.date._as_dict_
            if self.place:
                type_as_dict['place'] = self.place._as_dict_
            if self.value:
                type_as_dict['value'] = self.value
            if self.qualifiers and self.qualifiers != []:
                type_as_dict['qualifiers'] = [getattr(q, 'value', q) for q in self.qualifiers]
            log.debug(f"'Fact' serialized with fields: {type_as_dict.keys()}") 
            if type_as_dict == {}: log.warning("serializing and empty 'Fact'")
        
        return type_as_dict if type_as_dict != {} else None 
        
    @classmethod
    def _from_json_(cls, data: Dict[str, Any], context: Any = None) -> "Fact":
        if not isinstance(data, dict):
            raise TypeError(f"{cls.__name__}._from_json_ expected dict, got {type(data)}")
        fact_data = Conclusion._dict_from_json_(data,context)
        if (val := data.get("value")) is not None:
            fact_data["value"] = val
        if (date := data.get("date")) is not None:
            fact_data["date"] = Date._from_json_(date, context)
        if (place := data.get("place")) is not None:
            fact_data["place"] = PlaceReference._from_json_(place, context)
        if (ft := data.get("type")) is not None:
            fact_data["type"] = FactType(ft)
        if (quals := data.get("qualifiers")) is not None:
            fact_data["qualifiers"] = [Qualifier._from_json_(q) for q in quals if q is not None] 
            
        return cls(**fact_data) 

        







#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import html
import os
from typing import List, Optional, Tuple, Any
import re
from collections import defaultdict
from typing import Iterable, Iterator, List, Optional, Tuple, Union

import logging
from .logging_hub import hub, ChannelConfig
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================

job_id = "gedcomx.parsing.GEDCOM5x"

hub.start_channel(
    ChannelConfig(
        name=job_id,
        path=f"logs/{job_id}.log",
        level=logging.DEBUG,
        rotation="size:10MB:3",   # rotate by size, keep 3 backups
    )
)

log = logging.getLogger("gedcomx")

BOM = '\ufeff'

GEDCOM7_LINE_RE = re.compile(
    r"""^
    (?P<level>\d+)                    # Level
    (?:\s+@(?P<xref>[^@]+)@)?         # Optional record identifier
    \s+(?P<tag>[A-Z0-9_-]+)           # Tag
    (?:\s+(?P<value>.+))?             # Optional value (may be XREF)
    $""",
    re.VERBOSE
)

XREF_RE = re.compile(r'^@[^@]+@$')

# Add hash table for XREF of Zero Recrods?

nonzero = '[1-9]'
level = f'(?P<level>0|{nonzero}[0-9]*)'
atsign = '@'
underscore = '_'
ucletter = '[A-Z]'
tagchar = f'({ucletter}|[0-9]|{underscore})'
xref = f'{atsign}({tagchar})+{atsign}'
d = '\\ '
stdtag = f'{ucletter}({tagchar})*'
exttag = f'{underscore}({tagchar})+'
tag = f'({stdtag}|{exttag})'
voidptr = '@VOID@'
pointer = f'(?P<pointer>{voidptr}|{xref})'
nonat = '[\t -?A-\\U0010ffff]'
noneol = '[\t -\\U0010ffff]'
linestr = f'(?P<linestr>({nonat}|{atsign}{atsign})({noneol})*)'
lineval = f'({pointer}|{linestr})'
eol = '(\\\r(\\\n)?|\\\n)'
line = f'{level}{d}((?P<xref>{xref}){d})?(?P<tag>{tag})({d}{lineval})?{eol}'

from typing import List, Optional, Iterator, Union


class Gedcom5xRecord():
    def __init__(self,
                    line_num: Optional[int] = None,
                    level: int = -1,
                    tag: str | None = "NONR",
                    xref: Optional[str] = None,
                    value: Optional[str] = None,
                ) -> None:
        self.line = line_num
        self._subRecords: List[Gedcom5xRecord] = []
        self.level = int(level)
        self.xref = xref.replace('@','') if xref else ''
        self.pointer: bool = False
        self.tag = str(tag).strip()
        self.value = value

        self.parent: Optional[Gedcom5xRecord] = None
        self.root: Optional[Gedcom5xRecord] = None

    # ───────────────────────────────
    # Dict/JSON friendly view
    # ───────────────────────────────
    @property
    def _as_dict_(self):
        return {
            "level": self.level,
            "xref": self.xref,
            "tag": self.tag,
            "pointer": self.pointer,
            "value": self.value,
            "subrecords": [sub._as_dict_ for sub in self._subRecords],
        }

    # ───────────────────────────────
    # Subrecord management
    # ───────────────────────────────
    def addSubRecord(self, record: "Gedcom5xRecord"):
       
        if record is not None and (record.level == (self.level + 1)):
            record.parent = self
            self._subRecords.append(record)
        else:
            raise ValueError(
                f"SubRecord must be next level from this record (level:{self.level}, subRecord has level {record.level})"
            )

    def recordOnly(self):
        return Gedcom5xRecord(
            line_num=self.line, level=self.level, tag=self.tag, value=self.value
        )

    # ───────────────────────────────
    # Pretty printers
    # ───────────────────────────────
    def dump(self) -> str:
        record_dump = (
            f"Level: {self.level}, tag: {self.tag}, value: {self.value}, "
            f"subRecords: {len(self._subRecords)}\n"
        )
        for record in self._subRecords:
            record_dump += "\t" + record.dump()
        return record_dump

    def describe(self, subRecords: bool = False) -> str:
        level_str = "\t" * self.level
        description = (
            f"Line {self.line}: {level_str} Level: {self.level}, "
            f"tag: '{self.tag}', xref={self.xref} value: '{self.value}', "
            f"subRecords: {len(self._subRecords)}"
        )
        if subRecords:
            for subRecord in self.subRecords() or []:
                description += "\n" + subRecord.describe(subRecords=True)
        return description

    # ───────────────────────────────
    # Subrecord access
    # ───────────────────────────────
    def subRecord(self, tag: str):
        result = [r for r in self._subRecords if r.tag == tag]
        return None if not result else result

    def subRecords(self, tag: str | None = None) -> List['Gedcom5xRecord']:
        if not tag:
            return self._subRecords
        tags = tag.split("/", 1)

        # Collect matching first-level subrecords
        matches = [r for r in self._subRecords if r.tag == tags[0]]
        if not matches:
            return []

        if len(tags) == 1:
            return matches

        # Recurse deeper
        results = []
        for r in matches:
            sub_result = r.subRecords(tags[1])
            if sub_result:
                if isinstance(sub_result, list):
                    results.extend(sub_result)
                else:
                    results.append(sub_result)
        return results if results else []

    # ───────────────────────────────
    # Iteration / Subscriptability
    # ───────────────────────────────
    def __call__(self) -> str:
        return self.describe()

    def __iter__(self) -> Iterator["Gedcom5xRecord"]:
        """Iterates recursively over self and all subrecords."""
        yield from self._flatten_subrecords(self)

    def _flatten_subrecords(self, record: "Gedcom5xRecord") -> Iterator["Gedcom5xRecord"]:
        yield record
        for sub in record._subRecords:
            yield from self._flatten_subrecords(sub)

    def __len__(self) -> int:
        return len(self._subRecords)

    def __getitem__(self, key: Union[int, slice, str]) -> Union["Gedcom5xRecord", List["Gedcom5xRecord"]]:
        """
        - rec[0] -> first subrecord
        - rec[1:3] -> slice of subrecords
        - rec['NAME'] -> list of subrecords with tag 'NAME'
        """
        if isinstance(key, int) or isinstance(key, slice):
            return self._subRecords[key]
        if isinstance(key, str):
            matches = [r for r in self._subRecords if r.tag == key]
            if not matches:
                raise KeyError(f"No subrecords with tag '{key}'.")
            return matches[0] if len(matches) == 1 else matches
        raise TypeError(f"Unsupported key type: {type(key).__name__}")

    def __contains__(self, key: object) -> bool:
        if isinstance(key, str):
            return any(r.tag == key for r in self._subRecords)
        if isinstance(key, int):
            return 0 <= key < len(self._subRecords)
        return False


TagKey = str
IndexKey = int
Key = Union[IndexKey, slice, TagKey]

class Gedcom5x():
    """
    Object representing a Genealogy in legacy GEDCOM 5.x / 7 format. 

    Parameters
    ----------
    records : List[GedcomReord]
        List of GedcomRecords to initialize the genealogy with
    filepath : str
        path to a GEDCOM (``*``.ged), if provided object will read, parse and initialize with records in the file.
    
    Note
    ----
    **file_path** takes precidence over **records**.
    If no arguments are provided, Gedcom Object will initialize with no records.
    
    """
    _top_level_tags = ['INDI', 'FAM', 'OBJE', 'SOUR', 'REPO', 'NOTE', 'HEAD','SNOTE']
  
    def __init__(self, records: Optional[List[Gedcom5xRecord]] = None,filepath: str | None = None) -> None:

        self.records: List[Gedcom5xRecord] = records or []
        if filepath:
            self.records = self._records_from_file(filepath)
        elif records:
            self.records: List[Gedcom5xRecord] = records if records else []
        
        
        # Fast tag index: {'HEAD': [rec], 'INDI': [rec1, rec2, ...], ...}
        self._tag_index: dict[str, List[Gedcom5xRecord]] = defaultdict(list)
        self._reindex()
        
        self.header: Gedcom5xRecord | None = None
        self._sources: List[Gedcom5xRecord] = []
        self._repositories: List[Gedcom5xRecord] = []
        self._individuals: List[Gedcom5xRecord] = []
        self._families: List[Gedcom5xRecord] = []
        self._objects: List[Gedcom5xRecord] = []
        self._snotes: List[Gedcom5xRecord] = []
        self.version = None

        if self.records:
            for record in self.records:
                if record.tag == 'HEAD':
                    self.header = record
                    self.version = record['GEDC']['VERS'].value
                if record.tag == 'INDI':
                    self._individuals.append(record)
                if record.tag == 'SOUR' and record.level == 0:
                    self._sources.append(record)
                if record.tag == 'REPO' and record.level == 0:
                    self._repositories.append(record)
                if record.tag == 'FAM' and record.level == 0:
                    self._families.append(record)
                if record.tag == 'OBJE' and record.level == 0:
                    self._objects.append(record)
                if record.tag == 'SNOTE' and record.level == 0:
                    record.xref = record.value
                    self._snotes.append(record)

    # ─────────────────────────────────────────────────────────────
    # Subscriptable & iterable behavior
    # ─────────────────────────────────────────────────────────────
    def _reindex(self) -> None:
        """Rebuild the tag index from self.records."""
        self._tag_index.clear()
        for rec in self.records:
            # Normalize tag just in case
            tag = rec.tag if isinstance(rec.tag, str) else str(rec.tag)
            self._tag_index[tag].append(rec)

    def __len__(self) -> int:
        return len(self.records)

    def __iter__(self) -> Iterator['Gedcom5xRecord']:
        # Enables: for x in gedcom:
        return iter(self.records)

    def __contains__(self, key: object) -> bool:
        # Enables: 'HEAD' in gedcom  (tag membership)
        if isinstance(key, str):
            return key in self._tag_index and len(self._tag_index[key]) > 0
        if isinstance(key, int):
            return 0 <= key < len(self.records)
        return False

    def __getitem__(self, key: Key) -> Union['Gedcom5xRecord', List['Gedcom5xRecord']]:
        """
        - gedcom[0] -> GedcomRecord at index 0
        - gedcom[1:5] -> list of GedcomRecord (slice)
        - gedcom['HEAD'] -> single record if exactly one; otherwise list of matching records
        - gedcom['INDI'] -> list of all INDI records (usually many)
        """
        if isinstance(key, int):
            return self.records[key]
        if isinstance(key, slice):
            return self.records[key]
        if isinstance(key, str):
            matches = self._tag_index.get(key, [])
            if not matches:
                raise KeyError(f"No records with tag '{key}'.")
            # If exactly one match (e.g., HEAD), return the record; otherwise return list
            return matches[0] if len(matches) == 1 else matches
        raise TypeError(f"Unsupported key type: {type(key).__name__}")

    # Optional: convenience helpers
    def by_tag(self, tag: str) -> List['Gedcom5xRecord']:
        """Always return a list of records for a tag (empty list if none)."""
        return list(self._tag_index.get(tag, []))

    def first(self, tag: str) -> Optional['Gedcom5xRecord']:
        """Return the first record with a given tag, or None."""
        lst = self._tag_index.get(tag, [])
        return lst[0] if lst else None

    # If you add/replace records after init, keep the index fresh:
    def append(self, rec: 'Gedcom5xRecord') -> None:
        self.records.append(rec)
        self._tag_index.setdefault(rec.tag, []).append(rec)

    def extend(self, recs: Iterable['Gedcom5xRecord']) -> None:
        self.records.extend(recs)
        for r in recs:
            self._tag_index.setdefault(r.tag, []).append(r)

    def insert(self, idx: int, rec: 'Gedcom5xRecord') -> None:
        self.records.insert(idx, rec)
        self._tag_index.setdefault(rec.tag, []).append(rec)

    def remove(self, rec: 'Gedcom5xRecord') -> None:
        self.records.remove(rec)
        try:
            bucket = self._tag_index.get(rec.tag)
            if bucket:
                bucket.remove(rec)
                if not bucket:
                    del self._tag_index[rec.tag]
        except ValueError:
            pass  # already out of index

    def clear(self) -> None:
        self.records.clear()
        self._tag_index.clear()      
    # =========================================================
    # 2. PROPERTY ACCESSORS (GETTERS & SETTERS)
    # =========================================================
    
    @property
    def json(self):
        import json
        return json.dumps({'Individuals': [indi._as_dict_ for indi in self._individuals]},indent=4)

    @property
    def contents(self):
        def print_table(pairs):

            # Calculate the width of the columns
            name_width = max(len(name) for name, _ in pairs)
            value_width = max(len(str(value)) for _, value in pairs)

            # Print the header
            print(f'GEDCOM {self.version} Import Results')
            header = f"{'Type'.ljust(name_width)} | {'Count'.ljust(value_width)}"
            print('-' * len(header))
            print(header)
            print('-' * len(header))

            # Print each pair in the table
            for name, value in pairs:
                print(f"{name.ljust(name_width)} | {str(value).ljust(value_width)}")
                
        imports_stats = {
            'Top Level Records': len(self.records),
            'Individuals': len(self.individuals),
            'Family Group Records': len(self.families),
            'Repositories': len(self.repositories),
            'Sources': len(self.sources),
            'Objects': len(self.objects)
        }

        return imports_stats

    @property
    def sources(self) -> List[Gedcom5xRecord]:
        return self._sources

    @sources.setter
    def sources(self, value: List[Gedcom5xRecord]):
        if not isinstance(value, list) or not all(isinstance(item, Gedcom5xRecord) for item in value):
            raise ValueError("sources must be a list of GedcomRecord objects.")
        self._sources = value

    @property
    def repositories(self) -> List[Gedcom5xRecord]:
        """
        List of **REPO** records found in the Genealogy
        """
        return self._repositories

    @repositories.setter
    def repositories(self, value: List[Gedcom5xRecord]):
        if not isinstance(value, list) or not all(isinstance(item, Gedcom5xRecord) for item in value):
            raise ValueError("repositories must be a list of GedcomRecord objects.")
        self._repositories = value

    @property
    def individuals(self) -> List[Gedcom5xRecord]:
        return self._individuals

    @individuals.setter
    def individuals(self, value: List[Gedcom5xRecord]):
        if not isinstance(value, list) or not all(isinstance(item, Gedcom5xRecord) for item in value):
            raise ValueError("individuals must be a list of GedcomRecord objects.")
        self._individuals = value

    @property
    def families(self) -> List[Gedcom5xRecord]:
        return self._families

    @families.setter
    def families(self, value: List[Gedcom5xRecord]):
        if not isinstance(value, list) or not all(isinstance(item, Gedcom5xRecord) for item in value):
            raise ValueError("families must be a list of GedcomRecord objects.")
        self._families = value

    @property
    def objects(self) -> List[Gedcom5xRecord]:
        return self._objects

    @objects.setter
    def objects(self, value: List[Gedcom5xRecord]):
        if not isinstance(value, list) or not all(isinstance(item, Gedcom5xRecord) for item in value):
            raise ValueError("objects must be a list of GedcomRecord objects.")
        self._objects = value

    def write(self) -> bool:
        """
        Method placeholder for writing GEDCOM files.
        
        Raises
        ------
        NotImplementedError
         writing to legacy GEDCOM file is not currently implimented.
        """
        raise NotImplementedError("Writing of GEDCOM files is not implemented.")  

    @staticmethod
    def _records_from_file(file_path: str) -> List[Gedcom5xRecord]:
        def parse_gedcom7_line(line: str) -> Optional[Tuple[int, Optional[str], str, Optional[str], Optional[str]]]:
            """
            Parse a GEDCOM 7 line into: level, xref_id (record), tag, value, xref_value (if value is an @X@)
            
            Returns:
                (level, xref_id, tag, value, xref_value)
            """
            match = GEDCOM7_LINE_RE.match(line.strip())
            if not match:
                return None

            level = int(match.group("level"))
            xref_id = match.group("xref")
            xref_id = xref_id.strip('@') if xref_id else None
            tag = match.group("tag")
            value = match.group("value")
            if value == 'None': value = None
            xref_value = value.strip("@") if value and XREF_RE.match(value.strip()) else None

            return level, xref_id, tag, value, xref_value
        extension = '.ged'

        if not os.path.exists(file_path):
            print(f"File does not exist: {file_path}")
            raise FileNotFoundError
        elif not file_path.lower().endswith(extension.lower()):
            print(f"File does not have the correct extension: {file_path}")
            raise Exception("File does not appear to be a GEDCOM")
        
        print("Reading from GEDCOM file")
        with open(file_path, 'r', encoding='utf-8') as file:
            lines = [line.strip() for line in file]

            records = []
            record_map: dict[int,Any] = {0: None, 1: None, 2: None, 3: None, 4: None, 5: None}
            
            for l, line in enumerate(lines):
                if line.startswith(BOM):
                    line = line.lstrip(BOM)
                line = html.unescape(line).replace('&quot;', '')

                if line.strip() == '':
                    continue

                level, tag, value = '', '', ''

                # Split the line into the first two columns and the rest
                parts = line.split(maxsplit=2)
                if len(parts) == 3:
                    level, col2, col3 = parts

                    if col3 in Gedcom5x._top_level_tags:
                        tag = col3
                        value = col2
                    else:
                        tag = col2
                        value = col3
                
                else:
                    level, tag = parts

                level, xref, tag, value, xref_value = parse_gedcom7_line(line) or tuple([None, None, None, None])
                
                
                if xref is None and xref_value is not None:
                    xref = xref_value
               # print(l, level, xref, tag, value, xref_value)

                if isinstance(level,int):    
                    level = int(level)
                else: raise ValueError(f"Record had a level of {level}") 

                new_record = Gedcom5xRecord(line_num=l + 1, level=level, tag=tag if tag else None, xref=xref,value=value)
                
                
                if level == 0:
                    records.append(new_record)
                else:
                    new_record.root = record_map[0]
                    new_record.parent = record_map[int(level) - 1]
                    record_map[int(level) - 1].addSubRecord(new_record)
                record_map[int(level)] = new_record
                with hub.use(job_id):
                    log.info(new_record.describe())
        
        
        return records if records else []

    @staticmethod
    def fromFile(file_path: str) -> 'Gedcom5x':
        """
        Static method to create a Gedcom object from a GEDCOM file.

        Args:
            filepath (str): The path to the GEDCOM file.

        Returns:
            Gedcom: An instance of the Gedcom class.
        """     
        records = Gedcom5x._records_from_file(file_path)
        
        gedcom = Gedcom5x(records=records)      

        return gedcom
    
    def load_file(self,file_path: str) -> None:
        records = Gedcom5x._records_from_file(file_path)
        if records:
            self.records.extend(records)
            for record in self.records:
                if record.tag == 'HEAD':
                    pass
                    #self.header = record
                    #version = record['GEDC']['VERS'].value
                    #if not str(version)[0:2] == str(self.version)[0:2]: #TODO Deal with no VERS
                    #    raise ValueError(f'Wrong Version Current: {str(version)[0:2]}, new file: {str(self.version)[0:2]}')
                if record.tag == 'INDI':
                    self._individuals.append(record)
                if record.tag == 'SOUR' and record.level == 0:
                    self._sources.append(record)
                if record.tag == 'REPO' and record.level == 0:
                    self._repositories.append(record)
                if record.tag == 'FAM' and record.level == 0:
                    self._families.append(record)
                if record.tag == 'OBJE' and record.level == 0:
                    self._objects.append(record)
                if record.tag == 'SNOTE' and record.level == 0:
                    record.xref = record.value
                    self._snotes.append(record)
        else:
            raise ValueError()





from typing import Optional

from .resource import Resource
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================

class OnlineAccount:
    identifier = 'http://gedcomx.org/v1/OnlineAccount'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self, serviceHomepage: Resource, accountName: str) -> None:
        pass



from datetime import datetime
from typing import Optional, Dict, Any

"""
======================================================================
 Project: Gedcom-X
 File:    Attribution.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-08-25
 Updated:
   - 2025-08-31: fixed _as_dict_ to deal with Resources and ignore empty fields
   - 2025-09-03: _from_json_ refactor

   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .agent import Agent
from .resource import Resource
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================


class Attribution:
    """Attribution Information for a Genealogy, Conclusion, Subject and child classes

    Args:
        contributor (Agent, optional):            Contributor to object being attributed.
        modified (timestamp, optional):           timestamp for when this record was modified.
        changeMessage (str, optional):            Birth date (YYYY-MM-DD).
        creator (Agent, optional):      Creator of object being attributed.
        created (timestamp, optional):            timestamp for when this record was created

    Raises:
        
    """
    identifier = 'http://gedcomx.org/v1/Attribution'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self,contributor: Optional[Agent | Resource] = None,
                 modified: Optional[datetime] = None,
                 changeMessage: Optional[str] = None,
                 creator: Optional[Agent | Resource] = None,
                 created: Optional[datetime] = None) -> None:
               
        self.contributor = contributor
        self.modified = modified
        self.changeMessage = changeMessage
        self.creator = creator
        self.created = created
        
    @property
    def _as_dict_(self) -> Dict[str, Any] | None:
        """
        Serialize Attribution to a JSON-ready dict, skipping None values.
        """
        with hub.use(serial_log):
            log.debug(f"Serializing 'Attribution'")
            type_as_dict: Dict[str, Any] = {}
            if self.contributor:
                type_as_dict['contributor'] = Resource(target=self.contributor)._as_dict_  
            if self.modified:     
                type_as_dict['modified'] = self.modified if self.modified else None
            if self.changeMessage:
                type_as_dict['changeMessage'] = self.changeMessage if self.changeMessage else None 
            if self.creator:
                type_as_dict['creator'] = Resource(target=self.creator)._as_dict_   
            if self.created:      
                type_as_dict['created'] = self.created if self.created else None
                    
            log.debug(f"'Attribution' serialized with fields: {type_as_dict.keys()}") 
            if type_as_dict == {}: log.warning("serializing and empty 'Attribution'")
        return type_as_dict if type_as_dict != {} else None

    @classmethod
    def _from_json_(cls, data: Dict[str, Any],context) -> 'Attribution':
        if not isinstance(data, dict):
            raise TypeError(f"{cls.__name__}._from_json_ expected dict or str, got {type(data)}")

        attribution_data: Dict[str, Any] = {}

        # contributor: Agent | Resource | URI string
        if (contrib := data.get("contributor")) is not None:
            attribution_data["contributor"] = Resource._from_json_(contrib, context)

        # creator: Agent | Resource | URI string
        if (creator := data.get("creator")) is not None:
            attribution_data["creator"] = Resource._from_json_(creator, context)

        # changeMessage: str
        if (cm := data.get("changeMessage")) is not None:
            attribution_data["changeMessage"] = cm

        # created/modified: datetime
        if (created := data.get("created")) is not None:
            attribution_data["created"] = created
        if (modified := data.get("modified")) is not None:
            attribution_data["modified"] = modified

        return cls(**attribution_data)



from enum import Enum
from typing import List,Optional

"""
======================================================================
 Project: Gedcom-X
 File:    Name.py
 Author:  David J. Cartwright
 Purpose: Python Object representation of GedcomX Name, NameType, NameForm, NamePart Types

 Created: 2025-08-25
 Updated:
   - 2025-08-31: _as_dict_ to only create entries in dict for fields that hold data
   - 2025-09-03: _from_json_ refactor
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
#======================================================================
from .attribution import Attribution
from .conclusion import Conclusion, ConfidenceLevel
from .date import Date
from .document import Document
from .Extensions.rs10.rsLink import _rsLinks
from .note import Note
from .resource import Resource
from .source_reference import SourceReference
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================


class NameType(Enum):
    BirthName = "http://gedcomx.org/BirthName"
    MarriedName = "http://gedcomx.org/MarriedName"
    AlsoKnownAs = "http://gedcomx.org/AlsoKnownAs"
    Nickname = "http://gedcomx.org/Nickname"
    AdoptiveName = "http://gedcomx.org/AdoptiveName"
    FormalName = "http://gedcomx.org/FormalName"
    ReligiousName = "http://gedcomx.org/ReligiousName"
    Other = "other"
    
    @property
    def description(self):
        descriptions = {
            NameType.BirthName: "Name given at birth.",
            NameType.MarriedName: "Name accepted at marriage.",
            NameType.AlsoKnownAs: "\"Also known as\" name.",
            NameType.Nickname: "Nickname.",
            NameType.AdoptiveName: "Name given at adoption.",
            NameType.FormalName: "A formal name, usually given to distinguish it from a name more commonly used.",
            NameType.ReligiousName: "A name given at a religious rite or ceremony."
        }
        return descriptions.get(self, "No description available.")

class NamePartQualifier(Enum):
    Title = "http://gedcomx.org/Title"
    Primary = "http://gedcomx.org/Primary"
    Secondary = "http://gedcomx.org/Secondary"
    Middle = "http://gedcomx.org/Middle"
    Familiar = "http://gedcomx.org/Familiar"
    Religious = "http://gedcomx.org/Religious"
    Family = "http://gedcomx.org/Family"
    Maiden = "http://gedcomx.org/Maiden"
    Patronymic = "http://gedcomx.org/Patronymic"
    Matronymic = "http://gedcomx.org/Matronymic"
    Geographic = "http://gedcomx.org/Geographic"
    Occupational = "http://gedcomx.org/Occupational"
    Characteristic = "http://gedcomx.org/Characteristic"
    Postnom = "http://gedcomx.org/Postnom"
    Particle = "http://gedcomx.org/Particle"
    RootName = "http://gedcomx.org/RootName"
    
    @property
    def description(self):
        descriptions = {
            NamePartQualifier.Title: "A designation for honorifics (e.g., Dr., Rev., His Majesty, Haji), ranks (e.g., Colonel, General), positions (e.g., Count, Chief), or other titles (e.g., PhD, MD). Name part qualifiers of type Title SHOULD NOT provide a value.",
            NamePartQualifier.Primary: "A designation for the most prominent name among names of that type (e.g., the primary given name). Name part qualifiers of type Primary SHOULD NOT provide a value.",
            NamePartQualifier.Secondary: "A designation for a name that is not primary in its importance among names of that type. Name part qualifiers of type Secondary SHOULD NOT provide a value.",
            NamePartQualifier.Middle: "Useful for cultures designating a middle name distinct from a given name and surname. Name part qualifiers of type Middle SHOULD NOT provide a value.",
            NamePartQualifier.Familiar: "A designation for one's familiar name. Name part qualifiers of type Familiar SHOULD NOT provide a value.",
            NamePartQualifier.Religious: "A name given for religious purposes. Name part qualifiers of type Religious SHOULD NOT provide a value.",
            NamePartQualifier.Family: "A name that associates a person with a group, such as a clan, tribe, or patriarchal hierarchy. Name part qualifiers of type Family SHOULD NOT provide a value.",
            NamePartQualifier.Maiden: "Original surname retained by women after adopting a new surname upon marriage. Name part qualifiers of type Maiden SHOULD NOT provide a value.",
            NamePartQualifier.Patronymic: "A name derived from a father or paternal ancestor. Name part qualifiers of type Patronymic SHOULD NOT provide a value.",
            NamePartQualifier.Matronymic: "A name derived from a mother or maternal ancestor. Name part qualifiers of type Matronymic SHOULD NOT provide a value.",
            NamePartQualifier.Geographic: "A name derived from associated geography. Name part qualifiers of type Geographic SHOULD NOT provide a value.",
            NamePartQualifier.Occupational: "A name derived from one's occupation. Name part qualifiers of type Occupational SHOULD NOT provide a value.",
            NamePartQualifier.Characteristic: "A name derived from a characteristic. Name part qualifiers of type Characteristic SHOULD NOT provide a value.",
            NamePartQualifier.Postnom: "A name mandated by law for populations in specific regions. Name part qualifiers of type Postnom SHOULD NOT provide a value.",
            NamePartQualifier.Particle: "A grammatical designation for articles, prepositions, conjunctions, and other words used as name parts. Name part qualifiers of type Particle SHOULD NOT provide a value.",
            NamePartQualifier.RootName: "The 'root' of a name part, as distinguished from prefixes or suffixes (e.g., the root of 'Wilkówna' is 'Wilk'). A RootName qualifier MUST provide a value property."
        }
        return descriptions.get(self, "No description available.")

class NamePartType(Enum):
    Prefix = "http://gedcomx.org/Prefix"
    Suffix = "http://gedcomx.org/Suffix"
    Given = "http://gedcomx.org/Given"
    Surname = "http://gedcomx.org/Surname"
    
    @property
    def description(self):
        descriptions = {
            NamePartType.Prefix: "A name prefix.",
            NamePartType.Suffix: "A name suffix.",
            NamePartType.Given: "A given name.",
            NamePartType.Surname: "A surname."
        }
        return descriptions.get(self, "No description available.")
    
class NamePart:
    """Used to model a portion of a full name
        including the terms that make up that portion. Some name parts may have qualifiers
        to provide additional semantic meaning to the name part (e.g., "given name" or "surname").

    Args:
        type (NamePartType | None): Classification of this component of the name
            (e.g., ``Given``, ``Surname``, ``Prefix``, ``Suffix``, ``Title``, ``Particle``).
        value (str): The textual value for this part, without surrounding
            punctuation (e.g., ``"John"``, ``"van"``, ``"III"``).
        qualifiers (list[NamePartQualifier] | None): Optional qualifiers that refine
            the meaning or usage of this part (e.g., language/script variants, initials).

    Examples:
        >>> from gedcomx.Name import *
        >>> typ = NamePartType.Given
        >>> given = "Moses"
        >>> q = NamePartQualifier.Primary
        >>> name = NamePart(type=typ,value=given,qualifiers=[q])
        >>> print(name)
        NamePart(type=Given, value='Moses', qualifiers=1)

    """
    identifier = 'http://gedcomx.org/v1/NamePart'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self,
                 type: Optional[NamePartType] = None,
                 value: Optional[str] = None,
                 qualifiers: Optional[List[NamePartQualifier]] = None) -> None:
        self.type = type
        self.value = value
        self.qualifiers = qualifiers if qualifiers else []
    
    @property
    def _as_dict_(self):
        from .serialization import Serialization
        type_as_dict = {}
        if self.type:
            type_as_dict['type'] = self.type.value
        if self.value:
            type_as_dict['value'] = self.value
        if self.qualifiers:
            type_as_dict['qualifiers'] = [q.value for q in self.qualifiers]
        return type_as_dict if type_as_dict != {} else None
    
    @classmethod
    def _from_json_(cls, data: dict, context=None) -> "NamePart":
        if not isinstance(data, dict):
            raise TypeError(f"{cls.__name__}._from_json_ expected dict, got {type(data)}")

        name_part = {}

        # Enum / type
        if (typ := data.get("type")) is not None:
            name_part["type"] = NamePartType(typ)

        # String value
        if (val := data.get("value")) is not None:
            name_part["value"] = val

        # List of qualifiers
        if (quals := data.get("qualifiers")) is not None:
            name_part["qualifiers"] = [NamePartQualifier(q) for q in quals]

        return cls(**name_part)

    def __eq__(self, other):
        if not isinstance(other, NamePart):
            return NotImplemented
        return (self.type == other.type and
                self.value == other.value and
                self.qualifiers == other.qualifiers)    

    def __str__(self) -> str:
        parts = []
        if self.type is not None:
            parts.append(f"type={getattr(self.type, 'name', str(self.type))}")
        if self.value is not None:
            parts.append(f"value={self.value!r}")
        if self.qualifiers:
            parts.append(f"qualifiers={len(self.qualifiers)}")
        return f"NamePart({', '.join(parts)})" if parts else "NamePart()"

    def __repr__(self) -> str:
        if self.type is not None:
            tcls = self.type.__class__.__name__
            tname = getattr(self.type, "name", str(self.type))
            tval = getattr(self.type, "value", self.type)
            type_repr = f"<{tcls}.{tname}: {tval!r}>"
        else:
            type_repr = "None"
        return (
            f"{self.__class__.__name__}("
            f"type={type_repr}, "
            f"value={self.value!r}, "
            f"qualifiers={self.qualifiers!r})"
        )

class NameForm:
    """A representation of a name (a "name form") 
        within a given cultural context, such as a given language and script.
        As names are captured (both in records or in applications), the terms
        in the name are sometimes classified by type. For example, a certificate
        of death might prompt for **"given name(s)"** and **"surname"**. The parts
        list can be used to represent the terms in the name that have been classified.

    Args:
        lang (str | None): BCP-47 language tag for this name form (e.g., "en").
        fullText (str | None): The full, unparsed name string as written in the source.
            If provided, the name SHOULD be rendered as it would normally be spoken in
            the applicable cultural context.
        parts (list[NamePart] | None): Ordered structured components of the name
            (e.g., Given, Surname). If provided, ``fullText`` may be omitted. If
            provided, the list SHOULD be ordered such that the parts are in the order
            they would normally be spoken in the applicable cultural context.

    """
    identifier = 'http://gedcomx.org/v1/NameForm'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self, lang: Optional[str] = None,
                 fullText: Optional[str] = None,
                 parts: Optional[List[NamePart]] = None) -> None:
        
        self.lang = lang
        self.fullText = fullText
        self.parts = parts if parts else []
    
    @property
    def _as_dict_(self):
        from .serialization import Serialization
        type_as_dict = {}
        if self.lang:
            type_as_dict['lang'] = self.lang
        if self.fullText:
            type_as_dict['fullText'] = self.fullText
        if self.parts:
            type_as_dict['parts'] = [part._as_dict_ for part in self.parts if part is not None]
        return type_as_dict if type_as_dict != {} else None
        return Serialization.serialize_dict(type_as_dict)
    
    @classmethod
    def _from_json_(cls, data: dict, context=None) -> "NameForm":
        if not isinstance(data, dict):
            raise TypeError(f"{cls.__name__}._from_json_ expected dict, got {type(data)}")

        name_form = {}

        # Scalars
        if (lang := data.get("lang")) is not None:
            name_form["lang"] = lang

        if (full := data.get("fullText")) is not None:
            name_form["fullText"] = full

        # List of parts
        if (parts := data.get("parts")) is not None:
            name_form["parts"] = [NamePart._from_json_(p, context) for p in parts if p]

        return cls(**name_form)
    
    def _fulltext_parts(self):
        pass

class Name(Conclusion):
    """**Defines a name of a person.**

            A Name is intended to represent a single variant of a person's name. 
            This means that nicknames, spelling variations, or other names 
            (often distinguishable by a name type) should be modeled with 
            separate instances of Name.

    .. admonition:: Advanced details
        :class: toggle
    
        The name forms of a name contain alternate representations of the name.
        A Name MUST contain at least one name form, presumably a representation
        of the name that is considered proper and well formed in the person's native,
        historical cultural context. Other name forms MAY be included, which can be
        used to represent this name in contexts where the native name form is not easily
        recognized and interpreted. Alternate forms are more likely in situations where
        conclusions are being analyzed across cultural context boundaries that have both
        language and writing script differences.
        

    Attributes:
        type (Optional[:class:`~gedcomx.NameType`]): Classification of the name
            (e.g., BirthName, AlsoKnownAs).
        nameForms (List[:class:`~gedcomx.NameForm`]): One or more structured
            representations of the name (full text and parts).
        date (Optional[:class:`~gedcomx.Date`]): Date context for this name
            (e.g., when the name was used or recorded).
    """
    identifier = 'http://gedcomx.org/v1/Name'
    version = 'http://gedcomx.org/conceptual-model/v1'

    @staticmethod
    def simple(text: str):
        """
        Takes a string and returns a GedcomX Name Object
        """
        if text:
            text = text.replace("/","")
            parts = text.rsplit(' ', 1)
        
            # Assign val1 and val2 based on the split
            given = parts[0] if len(parts) > 1 else ""
            surname = parts[1] if len(parts) > 1 else parts[0]
            
            # Remove any '/' characters from both val1 and val2
            #given = given.replace('/', '')
            #surname = surname.replace('/', '')

            given_name_part = NamePart(type = NamePartType.Given, value=given)
            surname_part = NamePart(type = NamePartType.Surname, value=surname)

            name_form = NameForm(fullText=text,parts=[given_name_part,surname_part])
            name = Name(type=NameType.BirthName,nameForms=[name_form])
        else:
            name = Name()
        return name

    def __init__(self, id: Optional[str] = None,
                 lang: Optional[str] = None,
                 sources: Optional[List[SourceReference]] = None,
                 analysis: Optional[Document |Resource] = None,
                 notes: Optional[List[Note]] = None,
                 confidence: Optional[ConfidenceLevel] = None,
                 attribution: Optional[Attribution] = None,
                 type: Optional[NameType] = None,
                 nameForms: Optional[List[NameForm]]= None,
                 date: Optional[Date] = None,
                 links: Optional[_rsLinks] = None) -> None:
        super().__init__(id, lang, sources, analysis, notes, confidence, attribution,links=links)
        self.type = type
        self.nameForms = nameForms if nameForms else []
        self.date = date
    
    def _add_name_part(self, namepart: NamePart):
        if namepart and isinstance(namepart, NamePart):
            for current_namepart in self.nameForms[0].parts:
                if namepart == current_namepart:
                    return False
            self.nameForms[0].parts.append(namepart)
    
    @property
    def _as_dict_(self):
        type_as_dict = super()._as_dict_ or {}
        if self.type:
            type_as_dict['type'] = getattr(self.type, 'value', self.type)
        if self.nameForms:
            type_as_dict['nameForms'] = [nf._as_dict_ for nf in self.nameForms if nf]
        if self.date:
            type_as_dict['date'] = self.date._as_dict_
        
        return type_as_dict if type_as_dict != {} else None

    
    @classmethod
    def _from_json_(cls, data: dict,context = None) -> "Name":
        """Build a Name from JSON-like dict."""
        name = Conclusion._dict_from_json_(data)
        
        # Enum
        if (typ := data.get("type")) is not None:
            name["type"] = NameType(typ)

        # List
        if (forms := data.get("nameForms")) is not None:
            name["nameForms"] = [NameForm._from_json_(f, context) for f in forms]

        # Object
        if (date := data.get("date")) is not None:
            name["date"] = Date._from_json_(date, context)
        
        return cls(**name)
    
    def __str__(self) -> str:
        """Return a human-readable string for the Name-like object."""
        return f"Name(id={self.id}, type={self.type}, forms={len(self.nameForms)}, date={self.date})"

    def __repr__(self) -> str:
        """Return an unambiguous string representation of the Name-like object."""
        return (
            f"{self.__class__.__name__}("
            f"id={self.id!r}, "
            f"lang={self.lang!r}, "
            f"sources={self.sources!r}, "
            f"analysis={self.analysis!r}, "
            f"notes={self.notes!r}, "
            f"confidence={self.confidence!r}, "
            f"attribution={self.attribution!r}, "
            f"type={self.type!r}, "
            f"nameForms={self.nameForms!r}, "
            f"date={self.date!r})"
    )

class QuickName():
    def __new__(cls,name: str) -> Name:
        obj = Name(nameForms=[NameForm(fullText=name)])
        return obj
    
def ensure_list(val):
    if val is None:
        return []
    return val if isinstance(val, list) else [val]
    










from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Dict, Iterator, Literal

"""
======================================================================
 Project: Gedcom-X
 File:    ExtensibleEnum.py
 Author:  David J. Cartwright
 Purpose: Create a class that can act like an enum but be extended by the user at runtime.

 Created: 2025-08-25
 Updated:
   - YYYY-MM-DD: <change description>
   
======================================================================
"""

@dataclass(frozen=True, slots=True)
class _EnumItem:
    """
    A single registered member of an :class:`ExtensibleEnum`.

    Each `_EnumItem` represents one (name, value) pair that belongs
    to a particular `ExtensibleEnum` subclass. Items are immutable
    once created.

    Attributes
    ----------
    owner : type
        The subclass of :class:`ExtensibleEnum` that owns this member
        (e.g., `Color`).
    name : str
        The symbolic name of the member (e.g., `"RED"`).
    value : Any
        The underlying value associated with the member (e.g., `"r"`).

    Notes
    -----
    - Equality is determined by object identity (not overridden).
    - Instances are hashable by default since the dataclass is frozen.
    - The `__repr__` and `__str__` provide user-friendly string forms.

    Examples
    --------
    >>> class Color(ExtensibleEnum): ...
    >>> red = Color.register("RED", "r")
    >>> repr(red)
    'Color.RED'
    >>> str(red)
    'RED'
    >>> red.value
    'r'
    """
    owner: type
    name: str
    value: Any
    def __repr__(self) -> str:  # print(...) shows "Color.RED"
        return f"{self.owner.__name__}.{self.name}"
    def __str__(self) -> str:
        return self.name

class _ExtEnumMeta(type):
    def __iter__(cls) -> Iterator[_EnumItem]:
        return iter(cls._members.values())
    def __contains__(cls, item: object) -> bool:
        return item in cls._members.values()
    # Support Color('RED') / Color(2)
    def __call__(cls, arg: Any, /, *, by: Literal["auto","name","value"]="auto") -> _EnumItem:
        if isinstance(arg, _EnumItem):
            if arg.owner is cls:
                return arg
            raise TypeError(f"{arg!r} is not a member of {cls.__name__}")
        if by == "name":
            return cls.get(str(arg))
        if by == "value":
            return cls.from_value(arg)
        if isinstance(arg, str) and arg in cls._members:
            return cls.get(arg)
        return cls.from_value(arg)

class ExtensibleEnum(metaclass=_ExtEnumMeta):
    """
    A lightweight, **runtime-extensible**, enum-like base class.

    Subclass this to create an enum whose members can be registered at runtime.
    Registered members are exposed as class attributes (e.g., `Color.RED`) and
    can be retrieved by name (`Color.get("RED")`) or by value
    (`Color.from_value("r")`). Square-bracket lookup (`Color["RED"]`) is also
    supported via ``__class_getitem__``.

    This is useful when:
      - The full set of enum values is not known until runtime (plugins, config).
      - You need attribute-style access (`Color.RED`) but want to add members
        dynamically and/or validate uniqueness of names/values.

    Notes
    -----
    - **Uniqueness:** Names and values are unique within a subclass.
    - **Per-subclass registry:** Each subclass has its own member registry.
    - **Thread safety:** Registration is **not** thread-safe. If multiple threads
      may register members, wrap `register()` calls in your own lock.
    - **Immutability:** Once registered, a member’s `name` and `value` are fixed.
      Re-registering the same `name` with the *same* `value` returns the existing
      item; a different value raises an error.

    Examples
    --------
    Define an extensible enum and register members:

    >>> class Color(ExtensibleEnum):
    ...     pass
    ...
    >>> Color.register("RED", "r")
    _EnumItem(owner=Color, name='RED', value='r')
    >>> Color.register("GREEN", "g")
    _EnumItem(owner=Color, name='GREEN', value='g')

    Access members:

    >>> Color.RED is Color.get("RED")
    True
    >>> Color["GREEN"] is Color.get("GREEN")
    True
    >>> Color.from_value("g") is Color.GREEN
    True
    >>> Color.names()
    ['RED', 'GREEN']

    Error cases:

    >>> Color.register("RED", "different")  # doctest: +IGNORE_EXCEPTION_DETAIL
    ValueError: name 'RED' already used with different value 'r'
    >>> Color.get("BLUE")                    # doctest: +IGNORE_EXCEPTION_DETAIL
    KeyError: Color has no member named 'BLUE'
    >>> Color.from_value("b")                # doctest: +IGNORE_EXCEPTION_DETAIL
    KeyError: Color has no member with value 'b'
    """
    """Runtime-extensible enum-like base."""
    _members: Dict[str, _EnumItem] = {}

    def __init_subclass__(cls, **kw):
        super().__init_subclass__(**kw)
        cls._members = {}  # fresh registry per subclass

    @classmethod
    def __class_getitem__(cls, key: str) -> _EnumItem:  # Color['RED']
        return cls.get(key)

    @classmethod
    def register(cls, name: str, value: Any) -> _EnumItem:
        if not isinstance(name, str) or not name.isidentifier():
            raise ValueError("name must be a valid identifier")
        if name in cls._members:
            item = cls._members[name]
            if item.value != value:
                raise ValueError(f"name {name!r} already used with different value {item.value!r}")
            return item
        if any(m.value == value for m in cls._members.values()):
            raise ValueError(f"value {value!r} already used")
        item = _EnumItem(owner=cls, name=name, value=value)
        cls._members[name] = item
        setattr(cls, name, item)  # enables Color.RED attribute
        return item

    @classmethod
    def names(cls) -> list[str]:
        return list(cls._members.keys())

    @classmethod
    def get(cls, name: str) -> _EnumItem:
        try:
            return cls._members[name]
        except KeyError as e:
            raise KeyError(f"{cls.__name__} has no member named {name!r}") from e

    @classmethod
    def from_value(cls, value: Any) -> _EnumItem:
        for m in cls._members.values():
            if m.value == value:
                return m
        raise KeyError(f"{cls.__name__} has no member with value {value!r}")




from .textvalue import TextValue

class TopLevelTypeCollection:
    def __init__(self):
        self.items = {}
        self.items_by_type = {}
        self._id_idx ={}     # Hash Table for searching for item by id
        self._name_idx = {}
        self._uri_idx = {}
        self.authority = 'NewGedcomX'
        self.len = 0
    

    def append(self, item) -> bool:
        if item._uri._authority == '' or item._uri._authority is None:
            item._uri._authority = self.authority
        if item._uri._path == '' or item._uri._path is None:
            item._uri._path = f'{item.__class__.__name__}s'

        self.items[item._uri] = item
        self._update_indexes(item)
        self.len += 1
    
    def _update_indexes(self, item):
        # Update the id index
        if hasattr(item, 'id'):
            self._id_idx[item.id] = item
        if hasattr(item, 'names'):
            for name in item.names:
                if isinstance(name,TextValue):
                    self._name_idx[name] = item
                else:
                    break
        if item.__class__.__name__ in self.items_by_type.keys():
            self.items_by_type[item.__class__.__name__].append(item)
        else:
            self.items_by_type[item.__class__.__name__] = [item]

    
    def _dump_uris(self):
        for uri in self.items.keys():
            print(uri._uri)
    
    def __len__(self):
        return self.len
        
        



from .schemas import SCHEMA
from .agent import Agent
from .address import Address
from .attribution import Attribution
from .conclusion import Conclusion
from .converter import GedcomConverter
from .coverage import Coverage
from .date import Date
from .document import Document
from .document import DocumentType
from .evidence_reference import EvidenceReference
from .extensible_enum import ExtensibleEnum
from .event import Event
from .event import EventType
from .event import EventRole
from .extensible import Extensible
from. extensible import _ExtraField
from .fact import Fact
from .fact import FactQualifier
from .fact import FactType
from .gedcom import Gedcom
from .gedcom5x import Gedcom5x, Gedcom5xRecord
from .gedcomx import GedcomX
from .gender import Gender, GenderType
from .group import Group, GroupRole
from .identifier import Identifier, IdentifierType, IdentifierList
from .Logging import get_logger
from .name import Name, NameForm, NamePart, NamePartType, NameType, NamePartQualifier
from .note import Note
from .online_account import OnlineAccount
from .person import Person, QuickPerson
from .place_description import PlaceDescription
from .place_reference import PlaceReference
from .qualifier import Qualifier
from .relationship import Relationship, RelationshipType
from .serialization import Serialization
from .source_citation import SourceCitation
from .source_description import SourceDescription
from .source_description import ResourceType
from .source_reference import SourceReference
from .subject import Subject
from .textvalue import TextValue
from .resource import Resource
from .uri import URI


from .Extensions.rs10.rsLink import rsLink

from .gedcom7.gedcom7 import Gedcom7, GedcomStructure
from .translation import g7toXtable






from typing import List, Optional, Dict, Any
from collections.abc import Iterator
import json
import secrets
import string
import json

"""
======================================================================
 Project: Gedcom-X
 File:    identifier.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-08-25
 Updated:
   - 2025-09-03: _from_json_ refactor 
   - 2025-09-04: fixe identifier and identifieList json deserialization
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .extensible_enum import _EnumItem
from .resource import Resource
from .uri import URI
from .extensible_enum import ExtensibleEnum
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================

def make_uid(length: int = 10, alphabet: str = string.ascii_letters + string.digits) -> str:
    """
    Generate a cryptographically secure alphanumeric UID.

    Args:
        length: Number of characters to generate (must be > 0).
        alphabet: Characters to choose from (default: A-Za-z0-9).

    Returns:
        A random string of `length` characters from `alphabet`.
    """
    if length <= 0:
        raise ValueError("length must be > 0")
    return ''.join(secrets.choice(alphabet) for _ in range(length)).upper()

class IdentifierType(ExtensibleEnum):
    pass

"""Enumeration of identifier types."""
IdentifierType.register("Primary", "http://gedcomx.org/Primary")
IdentifierType.register("Authority", "http://gedcomx.org/Authority")            
IdentifierType.register("Deprecated", "http://gedcomx.org/Deprecated")
IdentifierType.register("Persistent", "http://gedcomx.org/Persistent")
IdentifierType.register("External", "https://gedcom.io/terms/v7/EXID")
IdentifierType.register("Other", "user provided")
IdentifierType.register("ChildAndParentsRelationship","http://familysearch.org/v1/ChildAndParentsRelationship")
    
class Identifier:
    identifier = 'http://gedcomx.org/v1/Identifier'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self, value: Optional[List[URI]], type: Optional[IdentifierType] = IdentifierType.Primary) -> None: # type: ignore
        if not isinstance(value,list):
            value = [value] if value else []
        self.type = type
        self.values = value if value else []
    
    @property
    def _as_dict_(self):
        from .serialization import Serialization
        type_as_dict = {}
        if self.values:
            type_as_dict["value"] = None # [v._as_dict_ for v in self.values]
        if self.type:
            type_as_dict["type"] = None #getattr(self.type, "value", self.type)  # type: ignore[attr-defined]

        return Serialization.serialize_dict(type_as_dict)

    @classmethod
    def _from_json_(cls, data: Dict[str, Any]) -> 'Identifier | None':
        """
        Construct an Identifier from a dict parsed from JSON.
        """
        
        for key in data.keys():
            type = key
            value = data[key]
        uri_obj: Optional[Resource] = None
        # TODO DO THIS BETTER

        # Parse type
        raw_type = data.get('type')
        if raw_type is None:
            return None
        id_type: Optional[IdentifierType] = IdentifierType(raw_type) if raw_type else None
        return cls(value=value, type=id_type)

class IdentifierList:
    def __init__(self) -> None:
        # maps identifier-type (e.g., str or IdentifierType.value) -> list of values
        self.identifiers: dict[str, list] = {}

    # -------------------- hashing/uniqueness helpers --------------------
    def make_hashable(self, obj):
        """Convert any object into a hashable representation."""
        if isinstance(obj, dict):
            return tuple(sorted((k, self.make_hashable(v)) for k, v in obj.items()))
        elif isinstance(obj, (list, set, tuple)):
            return tuple(self.make_hashable(i) for i in obj)
        elif isinstance(obj, URI):
            return obj._as_dict_
        elif hasattr(obj, "_as_dict_"):
            d = getattr(obj, "_as_dict_")
            return tuple(sorted((k, self.make_hashable(v)) for k, v in d.items()))
        else:
            return obj

    def unique_list(self, items):
        """Return a list without duplicates, preserving order."""
        seen = set()
        result = []
        for item in items:
            h = self.make_hashable(item)
            if h not in seen:
                seen.add(h)
                result.append(item)
        return result

    # -------------------- public mutation API --------------------
    def append(self, identifier: "Identifier"):
        if isinstance(identifier, Identifier):
            self.add_identifier(identifier)
        else:
            raise ValueError("append expects an Identifier instance")
    
    def add_identifier(self, identifier: "Identifier"):
        """Add/merge an Identifier (which may contain multiple values)."""
        if not (identifier and isinstance(identifier, Identifier) and identifier.type):
            raise ValueError("The 'identifier' must be a valid Identifier instance with a type.")
        if not isinstance(identifier.type,_EnumItem):
            raise ValueError
        key = identifier.type.value if hasattr(identifier.type, "value") else str(identifier.type)
        existing = self.identifiers.get(key, [])
        merged = self.unique_list(list(existing) + list(identifier.values))
        self.identifiers[key] = merged

    # -------------------- queries --------------------
    def contains(self, identifier: "Identifier") -> bool:
        """Return True if any of the identifier's values are present under that type."""
        if not (identifier and isinstance(identifier, Identifier) and identifier.type):
            return False
        key = identifier.type.value if hasattr(identifier.type, "value") else str(identifier.type)
        if key not in self.identifiers:
            return False
        pool = self.identifiers[key]
        # treat values as a list on the incoming Identifier
        for v in getattr(identifier, "values", []):
            if any(self.make_hashable(v) == self.make_hashable(p) for p in pool):
                return True
        return False

    # -------------------- mapping-like dunder methods --------------------
    def __iter__(self) -> Iterator[str]:
        """Iterate over identifier *types* (keys)."""
        return iter(self.identifiers)

    def __len__(self) -> int:
        """Number of identifier types (keys)."""
        return len(self.identifiers)

    def __contains__(self, key) -> bool:
        """Check if a type key exists (accepts str or enum with .value)."""
        k = key.value if hasattr(key, "value") else str(key)
        return k in self.identifiers

    def __getitem__(self, key):
        """Lookup values by type key (accepts str or enum with .value)."""
        k = key.value if hasattr(key, "value") else str(key)
        return self.identifiers[k]

    # (optional) enable assignment via mapping syntax
    def __setitem__(self, key, values):
        """Set/replace the list of values for a type key."""
        k = key.value if hasattr(key, "value") else str(key)
        vals = values if isinstance(values, list) else [values]
        self.identifiers[k] = self.unique_list(vals)

    def __delitem__(self, key):
        k = key.value if hasattr(key, "value") else str(key)
        del self.identifiers[k]

    # -------------------- dict-style convenience --------------------
    def keys(self):
        return self.identifiers.keys()

    def values(self):
        return self.identifiers.values()

    def items(self):
        return self.identifiers.items()

    def iter_pairs(self) -> Iterator[tuple[str, object]]:
        """Flattened iterator over (type_key, value) pairs."""
        for k, vals in self.identifiers.items():
            for v in vals:
                yield (k, v)
    
    @classmethod
    def _from_json_(cls, data,context=None):
        if isinstance(data, dict):
            identifier_list = IdentifierList()
            for key, vals in data.items():
                # Accept both single value and list in JSON
                vals = [URI._from_json_(v) for v in vals]
                identifier_list.add_identifier(
                    Identifier(value=vals, type=IdentifierType(key))
                )
            return identifier_list
        else:
            raise ValueError("Data must be a dict of identifiers.")

    @property
    def _as_dict_(self):
        type_as_dict = {}
        for k in self.identifiers.keys():
            #print(k,self.identifiers[k],type(self.identifiers[k]))
            type_as_dict[k] = [i._as_dict_ for i in self.identifiers[k]]
        return type_as_dict if type_as_dict != {} else None

    def __repr__(self) -> str:
        return json.dumps(self._as_dict_, indent=4)

    def __str__(self) -> str:
        return json.dumps(self._as_dict_)








from enum import Enum
from typing import List, Optional
"""
======================================================================
 Project: Gedcom-X
 File:    group.py
 Author:  David J. Cartwright
 Purpose: 

 Created: 2025-08-25
 Updated:
   - 2025-09-01: Updating basic structure, identify TODO s 
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""
from .attribution import Attribution
from .conclusion import ConfidenceLevel
from .document import Document
from .date import Date
from .evidence_reference import EvidenceReference
from .identifier import Identifier
from .note import Note
from .place_reference import PlaceReference
from .source_reference import SourceReference
from .resource import Resource

from .textvalue import TextValue
from .subject import Subject
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================

class GroupRoleType(Enum):                              #TODO Impliment
    def __init__(self) -> None:
        super().__init__()

class GroupRole:                                        #TODO Impliment
    identifier = 'http://gedcomx.org/v1/GroupRole'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self, person: Resource,type: Optional[Enum], date: Optional[Date],details: Optional[str]) -> None:
        pass

class Group(Subject):                                   #TODO Impliment
    identifier = 'http://gedcomx.org/v1/Group'
    version = 'http://gedcomx.org/conceptual-model/v1'

    def __init__(self,
                 id: str | None, lang: str | None,
                 sources: List[SourceReference] | None,
                 analysis: Document | Resource | None, notes: List[Note] | None, confidence: ConfidenceLevel | None, attribution: Attribution | None, extracted: bool | None, evidence: List[EvidenceReference] | None, media: List[SourceReference] | None, identifiers: List[Identifier] | None,
                 names: List[TextValue],
                 date: Optional[Date],
                 place: Optional[PlaceReference],
                 roles: Optional[List[GroupRole]]) -> None:
        super().__init__(id, lang, sources, analysis, notes, confidence, attribution, extracted, evidence, media, identifiers)
        self.names = names if names else []
        self.date = date
        self.place = place
        self.roles = roles if roles else []


from __future__ import annotations
from typing import List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .source_description import SourceDescription

from .attribution import Attribution
from .qualifier import Qualifier

from .resource import Resource

from .uri import URI
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================

from collections.abc import Sized

class KnownSourceReference(Qualifier):
    CharacterRegion = "http://gedcomx.org/CharacterRegion"
    RectangleRegion = "http://gedcomx.org/RectangleRegion"
    TimeRegion = "http://gedcomx.org/TimeRegion"
    Page = "http://gedcomx.org/Page"
    
    @property
    def description(self):
        descriptions = {
            self.CharacterRegion: (
                "A region of text in a digital document, in the form of a,b where a is the index of the start "
                "character and b is the index of the end character. The meaning of this qualifier is undefined "
                "if the source being referenced is not a digital document."
            ),
            self.RectangleRegion: (
                "A rectangular region of a digital image. The value of the qualifier is interpreted as a series "
                "of four comma-separated numbers. If all numbers are less than 1, it is interpreted as x1,y1,x2,y2, "
                "representing percentage-based coordinates of the top-left and bottom-right corners. If any number is "
                "more than 1, it is interpreted as x,y,w,h where x and y are coordinates in pixels, and w and h are "
                "the width and height of the rectangle in pixels."
            ),
            self.TimeRegion: (
                "A region of time in a digital audio or video recording, in the form of a,b where a is the starting "
                "point in milliseconds and b is the ending point in milliseconds. This qualifier's meaning is undefined "
                "if the source is not a digital audio or video recording."
            ),
            self.Page: (
                "A single page in a multi-page document, represented as a 1-based integer. This always references the "
                "absolute page number, not any custom page number. This qualifier is undefined if the source is not a "
                "multi-page document."
            )
        }
        return descriptions.get(self, "No description available.")
    
class SourceReference:
    identifier = 'http://gedcomx.org/v1/SourceReference'
    version = 'http://gedcomx.org/conceptual-model/v1'
    
    def __init__(self,
                 description: SourceDescription | URI | None = None,
                 descriptionId: Optional[str] = None,
                 attribution: Optional[Attribution] = None,
                 qualifiers: Optional[List[Qualifier]] = None
                 ) -> None:
        
        self.description = description
        self.descriptionId = descriptionId
        self.attribution = attribution
        self.qualifiers = qualifiers if qualifiers and isinstance(qualifiers, list) else [] 

    def add_qualifier(self, qualifier: Qualifier):
        if isinstance(qualifier, (Qualifier,KnownSourceReference)):
            if self.qualifiers:
                #TODO Prevent Duplicates
                for current_qualifier in self.qualifiers:
                    if qualifier == current_qualifier:
                        return
            self.qualifiers.append(qualifier)
            return
        raise ValueError("The 'qualifier' must be type 'Qualifier' or 'KnownSourceReference', not " + str(type(qualifier))) 
    
    def append(self, text_to_add: str):
        if text_to_add and isinstance(text_to_add, str):
            if self.descriptionId is None:
                self.descriptionId = text_to_add
            else:
                self.descriptionId += text_to_add
        else:
            raise ValueError("The 'text_to_add' must be a non-empty string.")
    
    @property    
    def _as_dict_(self):
        with hub.use(serial_log):
            log.debug(f"Serializing 'SourceReference' with 'descriptionId': '{self.descriptionId}'")
            type_as_dict = {}
            if self.description is not None:
                type_as_dict['description']= URI(target=self.description)._as_dict_ if self.description is not None else None
            if self.descriptionId is not None:
                type_as_dict['descriptionId']= self.descriptionId.replace("\n"," ").replace("\r"," ") if self.descriptionId else None
            if self.attribution is not None:
                type_as_dict['attribution'] = self.attribution._as_dict_ if self.attribution else None
            if self.qualifiers is not None and self.qualifiers != []:
                type_as_dict['qualifiers'] = [qualifier.__as_dict__ for qualifier in self.qualifiers] if (self.qualifiers and len(self.qualifiers) > 0) else None
            log.debug(f"'SourceReference' serialized with fields: {type_as_dict.keys()}") 
            if type_as_dict == {}: log.warning("serializing and empty 'SourceReference'")  
        return type_as_dict if type_as_dict != {} else None
      
    @classmethod
    def _from_json_(cls, data: dict, context=None) -> "SourceReference":
        ref = {}

        # Scalars
        if (descriptionId := data.get("descriptionId")) is not None:
            ref["descriptionId"] = descriptionId

        # Objects (description could be URI or SourceDescription)
        if (description := data.get("description")) is not None:
            # if description is just a string, assume URI
            if isinstance(description, str):
                ref["description"] = URI(description)
            elif isinstance(description, dict):
                print(">>",description)
                ref["description"] = Resource._from_json_(description, context)
                assert False
        else:
            pass #TODO
            #print(ref["descriptionId"])

        if (attribution := data.get("attribution")) is not None:
            ref["attribution"] = Attribution._from_json_(attribution, context)

        if (qualifiers := data.get("qualifiers")) is not None:
            ref["qualifiers"] = [Qualifier._from_json_(q, context) for q in qualifiers]

        return cls(**ref)
    
        

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False  

        return (
            self.description.uri == other.description.uri
        )



import re

from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================

class Gedcom():
    def __init__(self) -> None:
        pass

    @staticmethod
    def read_gedcom_version(filepath: str) -> str | None:
        """
        Reads only the HEAD section of a GEDCOM file and returns the GEDCOM standard version.
        Looks specifically for HEAD → GEDC → VERS.
        
        Returns:
            str: GEDCOM version (e.g., "5.5.1" or "7.0.0"), or None if not found.
        """
        version = None
        inside_head = False
        inside_gedc = False

        with open(filepath, "r", encoding="utf-8") as f:
            for line in f:
                parts = line.strip().split(maxsplit=2)
                if not parts:
                    continue

                level = int(parts[0])
                tag = parts[1] if len(parts) > 1 else ""
                value = parts[2] if len(parts) > 2 else None

                # Enter HEAD
                if level == 0 and tag == "HEAD":
                    inside_head = True
                    continue

                # Leave HEAD block
                if inside_head and level == 0:
                    break

                # Inside HEAD, look for GEDC
                if inside_head and level == 1 and tag == "GEDC":
                    inside_gedc = True
                    continue

                # If we drop back to level 1 (but not GEDC), stop looking inside GEDC
                if inside_gedc and level == 1:
                    inside_gedc = False

                # Inside GEDC, look for VERS
                if inside_gedc and tag == "VERS":
                    version = value
                    break

        return version



import logging

def get_logger(name=None):
    logger = logging.getLogger(name)
    if not logger.handlers:
        logger.setLevel(logging.DEBUG)

        formatter = logging.Formatter('[%(asctime)s] %(levelname)s - %(name)s - %(message)s')

        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

        # Optional: file logging
        file_handler = logging.FileHandler(f"{name}.log", encoding="utf-8")
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger



from typing import Optional

"""
======================================================================
 Project: Gedcom-X
 File:    Resource.py
 Author:  David J. Cartwright
 Purpose: References TopLevel Types for Serialization

 Created: 2025-08-25
 Updated:
   - 2025-08-31: working on target=Resource and deserialization issues
   - 2025-09-03: _from_json_ refactor, arguments changed
   
======================================================================
"""

"""
======================================================================
GEDCOM Module Types
======================================================================
"""

from .uri import URI
from .logging_hub import hub, logging
"""
======================================================================
Logging
======================================================================
"""
log = logging.getLogger("gedcomx")
serial_log = "gedcomx.serialization"
#=====================================================================
    
class Resource:
    """
    Class used to track and resolve URIs and references between datastores.

    Parameters
    ----------
    
    Raises
    ------
    ValueError
        If `id` is not a valid UUID.
    """
    # TODO, Deal with a resouce being passed, as it may be unresolved.
    def __init__(self,resource: URI | None = None,resourceId: str | None = None, target: object = None) -> None:
        
        #if (resource is None) and (target is None): #TODO
        #    raise ValueError('Resource object must point to something.')

        self.resource = resource
        self.resourceId = resourceId
        self.Id = None

        self.type = None
        self.resolved = False
        self.target: object = target
        self.remote: bool | None = None    # is the resource pointed to persitent on a remote datastore?

        if target:
            if isinstance(target,Resource):
                self.resource = target.resource
                self.resourceId = target.resourceId
                self.target = target.target
            else:
                log.debug(f"Target of type: {type(target)}, {target}")
                if hasattr(target,'uri'):
                    self.resource = target.uri
                else:
                    self.resourceId = URI(fragment=target.id)
                self.resourceId = target.id
   
    @property
    def uri(self):
        return self.resource
    
    @property
    def _as_dict_(self):
        from .serialization import Serialization
        return Serialization.serialize(self)
        typ_as_dict = {}
        if self.resource is not None:
            typ_as_dict["resource"] = self.resource
        if self.resourceId is not None:
            typ_as_dict["resourceId"] = self.resourceId,
        
        return typ_as_dict or None

    
    
    @classmethod
    def _from_json_(cls, data: dict, context=None) -> "Resource":
        if not isinstance(data, dict):
            raise TypeError(f"{cls.__name__}._from_json_ expected dict, got {type(data)} {data}")
        resource = {}

        # Scalars
        if (res := data.get("resource")) is not None:
            resource["resource"] = res

        return cls(**resource)

    def __repr__(self) -> str:
        return f"Resource(resource={self.resource}, resourceId={self.resourceId}, target={self.target})"
    
    def __str__(self) -> str:
        return f"resource={self.resource}{f', resourceId={self.resourceId}' if self.resourceId else ''}"
    


